#  Copyright (c) 2017-2022 Reveal Energy Services, Inc
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# This file is part of Orchid and related technologies.
#

# Tutorial DOM walk-through

"""
This script illustrates navigating a loaded project using the Orchid* Python API.

(*Orchid is a mark of Reveal Energy Services, Inc)
"""

# 0.5 Import packages

# The only import needed for the Python API is `orchid` itself.
import orchid

# Import other packages to meet specific needs
from dataclasses import dataclass  # Used to format DOM object data for readability
import pprint  # Used to "pretty-print" complex data, for example, lists
import textwrap  # Help to format pretty printed text
import uuid  # Used to construct an object ID from its string representation


# These functions and classes exists simply to assist in executing this script.


def banner(banner_text):
    print(len(banner_text) * '=')
    print(banner_text)
    print(len(banner_text) * '=')
    print()


def pretty_print_with_header(items, header):
    header_text = f'{header} returns:'
    pretty_printed_text = (textwrap
                           .TextWrapper(initial_indent=2 * ' ', subsequent_indent=(2 + 1) * ' ')
                           .fill(f'{pprint.pformat(items)}'))
    text_to_print = f'{header_text}\n{pretty_printed_text}'
    print(text_to_print)
    print()


def wait_for_input():
    input('Press enter to continue...')
    print()


@dataclass
class AboutProjectObject:
    """Class used to conveniently and consistently print project object identifying information."""
    name: str
    display_name: str
    object_id: uuid.UUID

    def __init__(self, project_object):
        self.name = project_object.name
        self.display_name = project_object.display_name
        self.object_id = project_object.object_id


@dataclass
class AboutStage(AboutProjectObject):
    """Specialized class for consistently and conveniently printing stage information."""
    display_stage_number: int = 0
    display_name_with_well: str = ''

    def __init__(self, stage):
        super().__init__(stage)
        self.display_stage_number = stage.display_stage_number
        self.display_name_with_well = stage.display_name_with_well


def navigate_dom():
    # 1.0 Load the .ifrac project

    # The following code simply captures the configured location of the Orchid training data. It is not
    # needed to use the Orchid Python API itself, but it is used in this example to load a well-known
    # `.ifrac` file.
    orchid_training_data_path = orchid.training_data_path()

    project = orchid.load_project(str(orchid_training_data_path.joinpath('frankNstein_Bakken_UTM13_FEET.ifrac')))

    # Our project is now loaded in memory. An Orchid project has many collections of other items. For
    # example, a project has a collection of wells, and a well has a collection of stages.
    #
    # Each of these objects, for example, each well and each stage, is identified by a unique identifier (an
    # instance of `uuid.UUID`). However, these identifiers, in order to be unique, are **not** easily
    # remembered by people. Further, Orchid **does not** require that alternatives, like a well name or
    # display name, be unique. To allow for convenient searching, project objects like wells and stages are
    # kept in a `SearchableProjectObjects` collection. This class provides methods for searching for more
    # specific instances:
    #
    # - `find_by_object_id()` - Returns the matching object or `None` if no such object exists
    # - `find_by_name()` - Returns an **iterator** of matching objects (since more than one may match).
    # - `find_by_display_name()` - Returns an **iterator** of matching objects.
    #
    # It provides methods returning all valid values of these keys:
    #
    # - `all_object_ids()`
    # - `all_names()`
    # - `all_display_names()`
    #
    # Since `find_by_object_id()`, `find_by_name()` and `find_by_display_name()` do not exhaust the criteria
    # you might want to use to find objects of interest, we have included a more generic method, `find()`,
    # that takes a predicate (a callable) and returns an iterator over all objects for which the predicate
    # returns `True`.
    #
    # Finally, we provide the general ability to iterate over all project objects using two mechanisms. The
    # older mechanism is the `all_objects()` method that returns an iterator over all objects in the
    # collection.
    #
    # Beginning with version 2021.3.318 of the Orchid Python API, we have made `SearchableProjectObjects` an
    # iterator. This change allows one to use the very familiar Python idiom of:
    #
    # ```
    # for well in projects.wells():
    #     // processing
    #     for a single well
    # ```
    #
    # Additionally, one can use `SearchableProjectObjects` in list, dictionary and generator expressions. For
    # example, to create a dictionary of wells keyed by the well object ID (since ** only ** object ID is
    # guaranteed to be unique) one would write:
    #
    # ```
    # wells_by_id = {well.object_id: well for well in project.wells()}
    # ```
    #
    # Using `SearchableProjectObjects` as an iterator is very general, is easily understood, and is most
    # useful when exploring a project interactively at a REPL. However, we have seen that using this general
    # technique easily leads to duplicated code to search through the collection of project objects for the
    # specific object that you actually need for your application. Consequently, we encourage you to use the
    # query methods of `SearchableProjectObjects`, including the general `find(...)` methods, first. If those
    # do not meet your needs, please let us know by creating
    # [an issue](https://github.com/Reveal-Energy-Services/orchid-python-api/issues).

    # 2.0 Query well "keys"

    # Particularly during exploration of a project, you may not know the specific object in which you are
    # interested, but you know something about its name or its display name. The Orchid Python API provides
    # you with the `all_names()` and `all_display_names()` to iterate over those names.

    banner('Query well identifying information: name, display name and object ID')

    all_well_names = list(project.wells().all_names())
    pretty_print_with_header(all_well_names, 'list(project.wells().all_names())')

    all_well_display_names = list(project.wells().all_display_names())
    pretty_print_with_header(all_well_display_names, 'list(project.wells().all_display_names())')

    # 2.1 Query all object ids

    # For completeness, we provide the `all_object_ids()` to list all the object IDs.
    all_well_object_ids = list(project.wells().all_object_ids())
    pretty_print_with_header(all_well_object_ids, 'list(project.wells().all_display_names())')

    wait_for_input()

    # 3.0 Find well by "key"

    banner('Find a collection of wells using name and display name')

    # The method, `find_by_name()`, returns an iterable over wells.
    wells_of_interest_by_name = list(project.wells().find_by_name('Demo_1H'))
    pretty_print_with_header([AboutProjectObject(well) for well in wells_of_interest_by_name],
                             'list(project.wells().find_by_name("Demo_1H"))')

    # Similarly, the method, `find_by_display_name()`, returns an iterable over wells
    wells_of_interest_by_display_name = list(project.wells().find_by_display_name('Demo_2H'))
    pretty_print_with_header([AboutProjectObject(well) for well in wells_of_interest_by_name],
                             'list(project.wells().find_by_display_name("Demo_1H"))')

    wait_for_input()

    # Because `find_by_name()` and `find_by_display_name()` returns an **iterator**, one typically must handle
    # this method returning
    #
    # - An empty iterator
    # - An iterator with more than 1 item
    # - An iterator with a single item
    #
    # For example,
    banner('One must handle the three possibilities of zero, one or many wells')

    well_name_of_interest = 'Demo_3H'
    wells_of_interest_by_name = list(project.wells().find_by_name(well_name_of_interest))
    if len(wells_of_interest_by_name) == 0:
        print(f'No well in project with name, {well_name_of_interest}')
        print()
    elif len(wells_of_interest_by_name) > 1:
        print(f'Found {len(wells_of_interest_by_name)} wells in project with name,'
              f' {well_name_of_interest}')
        print()
    else:
        pretty_print_with_header(AboutProjectObject(wells_of_interest_by_name[0]),
                                 f'project.wells().find_by_name({well_name_of_interest}) '
                                 f'finds single well')

    # Another way to handle multiple wells found by `find_by_name()` is to use `assert` statements. This way
    # is suitable if any number of matches other than one is an error.

    banner('Or one must use `assert` if not finding a single well is an error')

    assert len(wells_of_interest_by_name) == 1, (f'Expected one well with name, {well_name_of_interest},'
                                                 f' but found {len(wells_of_interest_by_display_name)}')
    well_of_interest = wells_of_interest_by_name[0]
    pretty_print_with_header(f'Assert single well with name, {well_name_of_interest}, in project\n'
                             f'  {AboutProjectObject(wells_of_interest_by_name[0])}',
                             f'project.wells().find_by_name({well_name_of_interest})')

    wait_for_input()

    banner('But one can find a single well using object ID')

    # However, `find_by_object_id()` method returns either a well with the specified object ID or None.
    #
    # If a well with this object ID exists:
    object_id = '9fe727b0-5fd1-4240-b475-51c1363edb0d'
    well_of_interest_by_object_id = project.wells().find_by_object_id(uuid.UUID(object_id))
    pretty_print_with_header((AboutProjectObject(well_of_interest_by_object_id)
                              if well_of_interest_by_object_id is not None
                              else "No such object"),
                             f'project.wells().find_by_object_id(uuid.UUID({object_id})')

    # But if no well with this object ID exists:
    banner('Remember that searching by object ID may return `None`')
    object_id = '00000000-0000-0000-0000-000000000000'
    well_of_interest_by_object_id = project.wells().find_by_object_id(uuid.UUID(object_id))
    well_of_interest = (AboutProjectObject(well_of_interest_by_object_id)
                        if well_of_interest_by_object_id is not None
                        else "No such object.")
    pretty_print_with_header(well_of_interest, f'project.wells().find_by_object_id(uuid.UUID({object_id})')

    wait_for_input()

    # 4.0 The `find()` method supports more generic queries

    banner('Use the "generic" `find` method with a predicate to find other wells of interest')

    # The `find()` method returns an iterable over the wells for which the specified predicate is `True`
    wells_of_interest = list(
        project.wells().find(lambda well: well.name == 'Demo_3H' or well.name == 'Demo_4H'))
    pretty_print_with_header([AboutProjectObject(well) for well in wells_of_interest],
                             'list(project.wells().find(<some condition(s)>))')

    wait_for_input()

    # 5.0 Finally, if you wish to iterate over all wells, use the `all_objects()` method

    banner('The Orchid Python API provides two mechanisms to iterate over items in `SearchableProjectObjects')

    banner('The `all_objects` method returns an iterator over all project objects in the collection')

    # The method, `all_objects()`, returns an iterable over **all** wells in the project
    wells_of_interest = list(project.wells().all_objects())
    pretty_print_with_header([AboutProjectObject(well) for well in wells_of_interest],
                             'list(project.wells().all_objects()')

    wait_for_input()

    banner('The `SearchableProjectObjects` class is an iterator over all project objects in the collection')

    # The `SearchableProjectObjects` class is an iterator and supports typical Python iterator idioms like:
    pretty_print_with_header([AboutProjectObject(well) for well in project.wells()],
                             'project.wells() returns a Python iterator')

    wait_for_input()

    # All the project top-level objects provide a similar interface:
    #
    # - `project.data_frames()`
    # - `project.monitors()`
    # - `project.time_series()`
    # - `project.wells()`
    #

    banner('All top-level project objects provide a similar interface')
    for top_level_name in ['Data frames',
                           'Monitors',
                           'Time series',
                           'Wells']:
        print(f'- {top_level_name}')
    print()

    wait_for_input()

    # Stages have the same interface; however, stages also have two additional methods:
    #
    # - `stage.find_by_display_stage_number()`
    # - `stage.find_by_display_name_with_well()`

    banner('Stages have two additional methods for searching')

    # 6.0 Additional `SearchableStages` methods

    # 6.1 Find stages with a specific display stage number

    banner('One can find a specific stage with a display stage number')

    # The method, `find_by_display_stage_number()`, returns either the single stage with the display
    # stage number or it returns `None`.
    #
    # Begin by searching for the well of interest.
    well_display_name_of_interest = 'Demo_1H'
    wells_of_interest_by_display_name = [
        well for well in project.wells().find_by_display_name(well_display_name_of_interest)
    ]

    assert len(wells_of_interest_by_display_name) == 1, \
        f'Expected one well with display_name, {well_display_name_of_interest},' \
        f' but found {len(wells_of_interest_by_display_name)}'
    well_of_interest = wells_of_interest_by_display_name[0]

    # Now search for a stage with a specified display stage number.
    stage_display_number_of_interest = 7
    stage_of_interest = well_of_interest.stages().find_by_display_stage_number(stage_display_number_of_interest)
    if stage_of_interest is not None:
        pretty_print_with_header(AboutStage(stage_of_interest),
                                 f'well_of_interest.stages().'
                                 f'find_by_display_stage_number({stage_display_number_of_interest})')
    else:
        print(f'No stage with display stage number, {stage_display_number_of_interest}.')

    banner('But again, searching by a display stage number may return `None`')

    # If you search for a stage by display stage number that does **not** exist:
    stage_display_number_of_interest = 9999
    stage_of_interest = well_of_interest.stages().find_by_display_stage_number(stage_display_number_of_interest)
    if stage_of_interest is not None:
        pretty_print_with_header(AboutStage(stage_of_interest),
                                 f'well_of_interest.stages().'
                                 f'find_by_display_stage_number({stage_display_number_of_interest})')
    else:
        print(f'No stage with display stage number, {stage_display_number_of_interest}.')

    wait_for_input()

    # 6.2 Find stages with a specific display name with well

    banner('One can find a collection of stages with a specified `display_name_with_well`')

    # The method, `find_by_display_name_with_well()`, returns:
    #
    # - An empty iterator if no stage with the specified display name with well exists
    # - An iterator with a single item if exactly one stage with the specified display name with well exists
    # - An iterator with multiple items if many stages with the specified display name with well exists
    #
    # For example, begin by searching for a well of interest.
    well_name_of_interest = 'Demo_2H'
    wells_of_interest_by_name = list(project.wells().find_by_name(well_name_of_interest))
    assert len(wells_of_interest_by_name) == 1, (f'Expected one well with name, {well_name_of_interest},'
                                                 f' but found {len(wells_of_interest_by_display_name)}')
    well_of_interest = wells_of_interest_by_name[0]

    stage_display_name_with_well_of_interest = 'Demo_2H-Stage-14'
    stages_of_interest = [AboutStage(stage) for stage
                          in well_of_interest.stages().find_by_display_name_with_well
                          (stage_display_name_with_well_of_interest)]
    pretty_print_with_header(stages_of_interest,
                             f'list(well_of_interest.stages().find_by_display_name_with_well'
                             f'({stage_display_name_with_well_of_interest})')

    banner('One must handle the three possibilities of zero, one or many stages')

    if len(stages_of_interest) == 0:
        print(f'No stage in well, {well_of_interest.display_name}, in project with name, '
              f'{stage_display_name_with_well_of_interest}')
        print()
    elif len(stages_of_interest) > 1:
        print(f'Found {len(stages_of_interest)} stages in well,'
              f' {well_of_interest.display_name}, with display name with well,'
              f' {stage_display_name_with_well_of_interest}')
        print(0)
    else:
        pretty_print_with_header(AboutStage(stages_of_interest[0]),
                                 f'project.wells().find_by_name({stage_display_name_with_well_of_interest}) '
                                 f'finds single stage')

    print()

    # Another way to handle multiple stages found by `find_by_display_name_with_well()` is to use `assert`
    # statements. This way is suitable if any number of matches other than one is an error.

    banner('Or one must use `assert` if not finding a single stage is an error')

    assert len(stages_of_interest) == 1, \
        (f'Expected one well with name, ' f'{stage_display_name_with_well_of_interest},'
         f' but found {len(stages_of_interest)}')
    stage_of_interest = stages_of_interest[0]

    print(AboutStage(stage_of_interest))

    wait_for_input()


def main():
    print(__doc__)
    navigate_dom()


if __name__ == '__main__':
    main()



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#

"""
This notebook illustrates using the Orchid* Python API and the `pandas` package to
search for and use project data frames.

(*Orchid is a mark of Reveal Energy Services, Inc)
"""

# Example: Searching for project data frames

# 0.5 Import packages

# The only import needed for the Orchid Python API is `orchid` itself.

import orchid

# The remaining imports are standard python packages to support the analysis.

import pprint
import uuid

import matplotlib.pyplot as plt
import pandas as pd
# The following import is included for its "side-effects" of an improved color schemes and
# plot styles. (See the "Tip" in section 9.2 of "Python for Data Analysis" for details.)
import seaborn as sns


def wait_for_input():
    input('Press enter to continue...')
    print()


def print_elided_data_frame(data_frame, title):
    print(title)
    print(data_frame.to_string(max_rows=10, max_cols=6))
    wait_for_input()


def search_data_frames():
    # 1.0 Load the .ifrac project

    # The following code simply captures the configured location of the Orchid training data. It is not needed to
    # use the Orchid Python API itself, but it is used in this example to load well-known data.

    orchid_training_data_path = orchid.training_data_path()

    print('Wait patiently for the project to load...')
    print()
    project = orchid.load_project(str(orchid_training_data_path.joinpath(
        'Project-frankNstein_Permian_UTM13FT_DF_PR2298_vs263.ifrac')))

    # 2.0 Search for a data frame using its object ID

    # I have copied to the object ID from Orchid itself.
    project_data_frame = project.data_frames().find_by_object_id(
        uuid.UUID('c08e6988-d8f5-4d7b-bccd-de968a5b398b'))
    if project_data_frame is not None:
        print_elided_data_frame(project_data_frame.pandas_data_frame(), 'Project Data Frame')

    # 3.0 Search for a data frame by its name

    fdi_data_frames = list(project.data_frames().find_by_name('FDI Observations'))
    assert(len(fdi_data_frames) == 1)
    fdi_data_frame = fdi_data_frames[0]
    print_elided_data_frame(fdi_data_frame.pandas_data_frame(), 'FDI Observations')

    # 4.0 Search for a data frame by its display name

    microseismic_data_frames = list(project.data_frames().find_by_display_name(
        'Microseismic Data Frame 01 (Potentially Corrupted)'))
    assert(len(microseismic_data_frames) == 1)
    microseismic_data_frame = microseismic_data_frames[0]
    print_elided_data_frame(microseismic_data_frame.pandas_data_frame(), 'Microseismic Data Frame')

    # 5.0 Query for all identifying information

    print('All data frame object IDs')
    pprint.pprint(list(project.data_frames().all_object_ids()))
    wait_for_input()

    print('All data frame names')
    pprint.pprint(list(project.data_frames().all_names()))
    wait_for_input()

    print('All data frame display names')
    pprint.pprint(list(project.data_frames().all_display_names()))
    wait_for_input()

    # 7.0 Leverage the full power of pandas

    stage_12_data_frame = project.data_frames().find_by_object_id(
        uuid.UUID('5304d2ac-dbf8-44db-8dd8-c2203714c456'))
    print_elided_data_frame(stage_12_data_frame.pandas_data_frame(), 'Stage 12 Data Frame')

    # Extract a specific column (the cumulative slurry of a stage)

    print('Cumulative Slurry (bbl)')
    print(stage_12_data_frame.pandas_data_frame()[' cum slurry [bll]'].to_string(max_rows=10))
    wait_for_input()

    # Find the maximum value (the volume to pick from the FDI data frame)

    print('Maximum Volume to Pick')
    print(max(fdi_data_frame.pandas_data_frame()['VolumeToPick']))
    wait_for_input()

    # 7.1 Summarize the completion data

    raw_completion_column_names = ['Timestamp', 'DeltaT', 'DeltaP', 'VolumeToPick', 'ProppantMass', 'Energy']

    # Extract completion columns into new data frame and change index to time
    raw_completion_details = fdi_data_frame.pandas_data_frame()[
        raw_completion_column_names].set_index('Timestamp')

    raw_completion_details.describe()

    # 7.2 Scale completion data to fit on single plot

    def scale_down_by_1000(v):
        return v / 1000.0

    copy_completion_details = raw_completion_details.copy()
    copy_completion_details['ProppantMass'] = copy_completion_details['ProppantMass'].transform(scale_down_by_1000)
    copy_completion_details['Energy'] = copy_completion_details['Energy'].transform(scale_down_by_1000)
    copy_completion_details['VolumeToPick'] = copy_completion_details['VolumeToPick'].transform(scale_down_by_1000)

    completion_details = copy_completion_details.rename({'ProppantMass': 'ProppantMass (1000 lbs)',
                                                         'Energy': 'Energy (1000 ft-lb)',
                                                         'VolumeToPick': 'VolumeToPick (Mbbl)'},
                                                        axis='columns', errors='raise')
    print_elided_data_frame(completion_details, 'Completion Data')

    # 7.3 Plot the completion data

    _axes = completion_details.plot(figsize=(8, 5), title='Completion Data')
    plt.show()


def main():
    print(__doc__)
    search_data_frames()


if __name__ == '__main__':
    main()



#  Copyright (c) 2017-2022 Reveal Energy Services, Inc 
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#

import deal

import numpy as np
import option
import toolz.curried as toolz

import orchid.base
from orchid import (
    dot_net_dom_access as dna,
    net_quantity as onq,
    reference_origins as origins,
    unit_system as units,
    validation,
)

# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics import IWellTrajectory

# noinspection PyUnresolvedReferences
import UnitsNet


def _trajectory_array_in_unit(tgt_unit, raw_array):
    result = toolz.pipe(raw_array,
                        toolz.map(option.maybe),
                        toolz.map(onq.as_measurement(tgt_unit)),
                        toolz.map(lambda m: m.magnitude),
                        lambda magnitudes: np.fromiter(magnitudes, dtype='float'))
    return result


class NativeTrajectoryAdapterIdentified(dna.IdentifiedDotNetAdapter):
    def __init__(self, net_trajectory: IWellTrajectory):
        """
        Construct an instance adapting a .NET `IWellTrajectory`.

        One need not construct an instance directly; instead, one constructs an instance by accessing the
        Orchid domain object model (DOM) of a project. For example,

        >>> project_path = orchid.training_data_path().joinpath('frankNstein_Bakken_UTM13_FEET.ifrac')
        >>> project = orchid.load_project(str(project_path))
        >>> well = list(project.wells().find_by_name('Demo_1H'))[0]
        >>> # noinspection PyUnresolvedReferences
        >>> trajectory = well.trajectory

        Args:
            net_trajectory: The .NET trajectory to be adapted.
        """
        super().__init__(net_trajectory, orchid.base.constantly(net_trajectory.Well.Project))

    # TODO: Consider alternative interface, get_eastings()
    # See [pint Numpy support](https://pint.readthedocs.io/en/stable/numpy.html) or the
    # [pint pandas extension](https://github.com/hgrecco/pint-pandas)
    @deal.pre(validation.arg_not_none)
    def get_easting_array(self, reference_frame: origins.WellReferenceFrameXy) -> np.array:
        """
        Calculate the magnitudes of eastings of this trajectory relative to the specified `reference_frame`
        in project length units.

        Once one has a trajectory (from a well of a project), one can access the (Numpy) array of easting
        values relative to a reference. For example,

        >>> project_path = orchid.training_data_path().joinpath('frankNstein_Bakken_UTM13_FEET.ifrac')
        >>> project = orchid.load_project(str(project_path))
        >>> well = list(project.wells().find_by_name('Demo_3H'))[0]
        >>> # noinspection PyUnresolvedReferences
        >>> trajectory = well.trajectory
        >>> eastings = trajectory.get_easting_array(origins.WellReferenceFrameXy.PROJECT)
        >>> eastings[213]
        -21007.733000000007

        Args
            reference_frame: The reference frame for the easting coordinates.

        Returns:
            The array of eastings of this trajectory.
        """
        return self._trajectory_length_array(self.dom_object.GetEastingArray(reference_frame))

    # TODO: Consider alternative interface, get_northings()
    # See [pint Numpy support](https://pint.readthedocs.io/en/stable/numpy.html) or the
    # [pint pandas extension](https://github.com/hgrecco/pint-pandas)
    @deal.pre(validation.arg_not_none)
    def get_northing_array(self, reference_frame: origins.WellReferenceFrameXy) -> np.array:
        """
        Calculate the magnitudes of northings of this trajectory relative to the specified `reference_frame`
        in project length units.

        Once one has a trajectory (from a well of a project), one can access the (Numpy) array of northing
        values relative to a reference. For example,

        >>> project_path = orchid.training_data_path().joinpath('frankNstein_Bakken_UTM13_FEET.ifrac')
        >>> project = orchid.load_project(str(project_path))
        >>> well = list(project.wells().find_by_name('Demo_2H'))[0]
        >>> # noinspection PyUnresolvedReferences
        >>> trajectory = well.trajectory
        >>> northings = trajectory.get_northing_array(origins.WellReferenceFrameXy.PROJECT)
        >>> northings[203]
        36105.4299999997

        Args:
            reference_frame: The reference frame for the northing coordinates.

        Returns:
            The array of northings of this trajectory.

        """
        return self._trajectory_length_array(self.dom_object.GetNorthingArray(reference_frame))

    # TODO: Consider alternative interface, get_tvd_sss()
    # See [pint Numpy support](https://pint.readthedocs.io/en/stable/numpy.html) or the
    # [pint pandas extension](https://github.com/hgrecco/pint-pandas)
    def get_tvd_ss_array(self) -> np.array:
        """
        Calculate the array of total vertical sub-sea depth values in project length units.

        Once one has a trajectory (from a well of a project), one can access the (Numpy) array of TVDSS
        values. For example,

        >>> project_path = orchid.training_data_path().joinpath('frankNstein_Bakken_UTM13_FEET.ifrac')
        >>> project = orchid.load_project(str(project_path))
        >>> well = list(project.wells().find_by_name('Demo_4H'))[0]
        >>> # noinspection PyUnresolvedReferences
        >>> trajectory = well.trajectory
        >>> tvd_sss = trajectory.get_tvd_ss_array()
        >>> tvd_sss[144]
        10763.48512

        Returns:
            The array of total vertical depths of this trajectory.
        """
        return self._get_tvd_array(origins.DepthDatum.SEA_LEVEL)

    # TODO: Consider alternative interface, get_inclinations()
    # See [pint Numpy support](https://pint.readthedocs.io/en/stable/numpy.html) or the
    # [pint pandas extension](https://github.com/hgrecco/pint-pandas)
    def get_inclination_array(self) -> np.array:
        """
        Calculate the array of inclination values.

        Once one has a trajectory (from a well of a project), one can access the (Numpy) array of inclination
        values. For example,

        >>> project_path = orchid.training_data_path().joinpath('Project-frankNstein_Montney_UTM13_METERS.ifrac')
        >>> project = orchid.load_project(str(project_path))
        >>> well = list(project.wells().find_by_name('Hori_03'))[0]
        >>> # noinspection PyUnresolvedReferences
        >>> trajectory = well.trajectory
        >>> inclinations = trajectory.get_inclination_array()
        >>> inclinations[10]
        4.2

        Returns:
            The array of inclinations of this trajectory.
        """
        return self._trajectory_angle_array(self.dom_object.GetInclinationArray())

    # TODO: Consider alternative interface, get_azimuths_east_of_north()
    # See [pint Numpy support](https://pint.readthedocs.io/en/stable/numpy.html) or the
    # [pint pandas extension](https://github.com/hgrecco/pint-pandas)
    def get_azimuth_east_of_north_array(self) -> np.array:
        """
        Calculate the array of azimuth values.

        Once one has a trajectory (from a well of a project), one can access the (Numpy) array of azimuth east
        of north values. For example,

        >>> project_path = orchid.training_data_path().joinpath('Project-frankNstein_Montney_UTM13_METERS.ifrac')
        >>> project = orchid.load_project(str(project_path))
        >>> well = list(project.wells().find_by_name('Hori_03'))[0]
        >>> # noinspection PyUnresolvedReferences
        >>> trajectory = well.trajectory
        >>> azimuths_east_of_north = trajectory.get_azimuth_east_of_north_array()
        >>> azimuths_east_of_north[173]
        128.6

        Returns:
            The array of azimuths of this trajectory.
        """
        return self._trajectory_angle_array(self.dom_object.GetAzimuthEastOfNorthArray())

    # TODO: Consider alternative interface, get_md_kbs()
    # See [pint Numpy support](https://pint.readthedocs.io/en/stable/numpy.html) or the
    # [pint pandas extension](https://github.com/hgrecco/pint-pandas)
    def get_md_kb_array(self) -> np.array:
        """
        Calculate the array of MD KB values.

        Once one has a trajectory (from a well of a project), one can access the (Numpy) array of MDKB
        values. For example,

        >>> project_path = orchid.training_data_path().joinpath('Project-frankNstein_Montney_UTM13_METERS.ifrac')
        >>> project = orchid.load_project(str(project_path))
        >>> well = list(project.wells().find_by_name('Hori_01'))[0]
        >>> # noinspection PyUnresolvedReferences
        >>> trajectory = well.trajectory
        >>> md_kbs = trajectory.get_md_kb_array()
        >>> md_kbs[84]
        2246.04

        Returns:
            The array of MD KB values of this trajectory.
        """
        return self._trajectory_length_array(self.dom_object.GetMdKbArray())

    @deal.pre(validation.arg_not_none)
    def _get_tvd_array(self, depth_datum: origins.DepthDatum) -> np.array:
        """
        Calculate the array of total vertical depth values relative to `depth_datum` in project length units.

        Args:
            depth_datum: The depth datum for the vertical depth values.

        Returns:
            The array of total vertical depths of this trajectory.
        """
        return self._trajectory_length_array(self.dom_object.GetTvdArray(depth_datum))

    @staticmethod
    def _trajectory_angle_array(raw_array):
        return _trajectory_array_in_unit(units.Common.ANGLE, raw_array)

    def _trajectory_length_array(self, raw_array):
        return _trajectory_array_in_unit(self.expect_project_units.LENGTH, raw_array)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#

import pendulum

import orchid.base
from orchid import (
    dot_net_dom_access as dna,
    dom_project_object as dpo,
    native_time_series_adapter as tsa,
    net_date_time as ndt,
)

# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics import IMonitor


class NativeMonitorAdapter(dpo.DomProjectObject):
    """Adapts a native IMonitor to python."""
    def __init__(self, net_monitor: IMonitor):
        """
        Constructs an instance adapting a .NET IMonitor.

        Args:
            net_monitor: The .NET monitor to be adapted.
        """
        super().__init__(net_monitor, orchid.base.constantly(net_monitor.Project))

    start_time = dna.transformed_dom_property('start_time', 'The start time of this monitor.',
                                              ndt.as_date_time)
    stop_time = dna.transformed_dom_property('stop_time', 'The stop time of this monitor.',
                                             ndt.as_date_time)
    well_time_series = dna.transformed_dom_property(
        'time_series',
        """The complete time series for the well that may be monitoring treatments.
        
        The returned time series contains all samples recorded for the project. More specifically, it may
        contain samples either before or after `time_range()`.
        
        Another consequence of this definition is that if a client wants to access sample inside 
        `time_range()`, one must write code to manually filter the series samples to only include the samples
        in `time_range()`.
        """,
        tsa.NativeTimeSeriesAdapter)

    @property
    def time_range(self):
        """
        Calculate the time range during which this monitor is active; that is, monitoring the treatment stage.

        Returns:
            The time range during which this monitor is active. The type of the returned value is
            `pendulum.Period`. See the [documentation](https://pendulum.eustace.io/docs/) to understand the
            methods available from a `pendulum.Period` instance.
        """
        return pendulum.Period(self.start_time, self.stop_time)



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#

from typing import Callable, Iterator, Optional
import uuid

import toolz.curried as toolz

from orchid import dom_project_object as dpo

# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics import IProjectObject


class SearchableProjectError(ValueError):
    """
    Raised when an error occurs searching for a `dpo.DomProjectObject`.
    """
    pass


class SearchableProjectMultipleMatchError(SearchableProjectError):
    """
    Raised when multiple matches occur when searching for a `dpo.DomProjectObject`.
    """
    pass


"""
Provides a searchable collection of `DomProjectObject` instances. This searchable collection provide methods to:

- Query for all object IDs identifying instances in this collection
- Query for all objects in this collection
- Query for the name of all instances in this collection
- Query for the display name of all instances in this collection
- Search for a single instance by object ID
- Search for all instances with a specified name
- Search for all instances with a specified display name

Here are the DOM objects that currently may be collections:
- Data frames
- Monitors
- Stages
- Stage Parts
- Well trajectory
- Wells

This objects are all derived from `IProjectObject`. The corresponding instances in the Python API all derive from 
`dpo.DomProjectObject` which implements the attributes `object_id`, `name` and `display_name`.
"""


class SearchableProjectObjects:
    def __init__(self, make_adapter: Callable, net_project_objects: Iterator[IProjectObject]):
        """
        Construct a collection of project objects created my `make_adapter` using the arguments, `net_project_objects`.
        Args:
            make_adapter: The callable that constructs adapter instances using `net_project_objects`.
            net_project_objects: The sequence of .NET `IProjectObject` instances adapted by the Python API.
        """
        def add_project_object_to_collection(so_far, update_po):
            return toolz.assoc(so_far, update_po.object_id, update_po)
        
        self._collection = toolz.pipe(
            net_project_objects,
            toolz.map(make_adapter),
            lambda pos: toolz.reduce(add_project_object_to_collection, pos, {}),
        )

    def __iter__(self):
        """
        Return an iterator over the items in this collection.

        Returns:
            An iterator over the items in this collection.
        """
        # TODO: Change this implementation to be more "fundamental" than `all_objects()`.
        # The implementation of this method currently calls `all_objects()`. I think this implementation
        # is not quite correct; instead, `all_objects()` should call this method (even if indirectly) which
        # should return an iterator over `self._collection.values()`.
        #
        # Because we have not tested the Orchid Python API much this quarter, because we are creating a
        # release, and because I do not want to destabilize the release at this time, I have chosen to
        # leave the implementation of `all_objects()` as is and implement this method in terms of
        # `all_objects()`. (FYI: I did prototype "switching" the implementations and all unit tests passed.)
        return iter(self.all_objects())

    def __len__(self):
        """
        Return the number of items in this collection.

        Returns:
            The number of items in this collection.
        """
        return len(self._collection)

    def all_display_names(self) -> Iterator[str]:
        """
        Return an iterator over all the display names of project objects in this collection.

        Returns:
            An iterator over all the display names of project objects in this collection.
        """
        return toolz.map(lambda po: po.display_name, self._collection.values())

    def all_names(self) -> Iterator[str]:
        """
        Return an iterator over all the names of project objects in this collection.

        Returns:
            An iterator over all the names of project objects in this collection.
        """
        return toolz.map(lambda po: po.name, self._collection.values())

    def all_object_ids(self) -> Iterator[uuid.UUID]:
        """
        Return an iterator over all the object IDs of project objects in this collection.

        Returns:
            An iterator over all the object IDs of project objects in this collection.
        """
        return self._collection.keys()

    def all_objects(self) -> Iterator[dpo.DomProjectObject]:
        """
        Return an iterator over all the projects objects in this collection.

        Returns:
            An iterator over all the project objects in this collection.
        """
        return self._collection.values()

    def find(self, predicate: Callable) -> Iterator[dpo.DomProjectObject]:
        """
        Return an iterator over all project objects for which `predicate` returns `True`.

        Args:
            predicate: The `boolean` `Callable` to be invoked for each `dpo.DomProjectObject` in the collection.

        Returns:
            An iterator over all project objects fulfilling `predicate`.
        """
        return toolz.filter(predicate, self._collection.values())

    def find_by_display_name(self, display_name_to_find: str) -> Iterator[dpo.DomProjectObject]:
        """
        Return an iterator over all project objects whose `display_name` is the `display_name_to_find`.

        Args:
            display_name_to_find: The display name for all project objects of interest.

        Returns:
            An iterator over all project objects with the specified `display_name` property.
        """
        return self.find(lambda po: po.display_name == display_name_to_find)

    def find_by_name(self, name_to_find: str) -> Iterator[dpo.DomProjectObject]:
        """
        Return an iterator over all project objects whose `name` is the `name_to_find`.

        Args:
            name_to_find: The name for all project objects of interest.

        Returns:
            An iterator over all project objects with the specified `name` property.
        """
        return self.find(lambda po: po.name == name_to_find)

    def find_by_object_id(self, object_id_to_find: uuid.UUID) -> Optional[dpo.DomProjectObject]:
        """
        Return the project object whose `object_id` is the `object_id_to_find` if available; otherwise, `None`.

        Args:
            object_id_to_find: The object ID for the project objects of interest.

        Returns:
            The project objects with the specified `name` property. If no such project is found, return `None`.
        """
        return toolz.get(object_id_to_find, self._collection, default=None)



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#

from orchid import searchable_project_objects as spo


class SearchableStages(spo.SearchableProjectObjects):
    def find_by_display_stage_number(self, to_find: int):
        candidates = list(self.find(lambda s: s.display_stage_number == to_find))
        if len(candidates) == 0:
            return None
        elif len(candidates) == 1:
            return candidates[0]
        else:
            raise spo.SearchableProjectMultipleMatchError(to_find)

    def find_by_display_name_with_well(self, to_find: str):
        return self.find(lambda s: s.display_name_with_well == to_find)



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#

import argparse
import functools
import logging
import pathlib

import clr
import orchid
from orchid import (
    dot_net_disposable as dnd,
    net_fracture_diagnostics_factory as net_factory,
)

# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics import (MonitorExtensions, Leakoff, Observation)
# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.Factories.Implementations import LeakoffCurves
# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.SDKFacade import (
    ScriptAdapter,
)
# noinspection PyUnresolvedReferences
from System import (Array, Double, DateTime, String)
# noinspection PyUnresolvedReferences
from System.IO import (FileStream, FileMode, FileAccess, FileShare)
# noinspection PyUnresolvedReferences
import UnitsNet

clr.AddReference('Orchid.Math')
clr.AddReference('System.Collections')
# noinspection PyUnresolvedReferences
from Orchid.Math import Interpolation
# noinspection PyUnresolvedReferences
from System.Collections.Generic import List


object_factory = net_factory.create()


def calculate_delta_pressure(leak_off_pressure, maximum_pressure_sample):
    """
    Calculate the delta pressure value.

    Args:
        leak_off_pressure: Pressure from the treatment leak off curve.
        maximum_pressure_sample: The maximum treatment pressure.

    Returns:
        The pressure difference.

    """
    return UnitsNet.Pressure.op_Subtraction(
        UnitsNet.Pressure(maximum_pressure_sample.Value, UnitsNet.Units.PressureUnit.PoundForcePerSquareInch),
        leak_off_pressure)


def calculate_leak_off_control_point_times(interpolation_point_1, interpolation_point_2, ticks):
    """
    Return the calculated control points for a leak off curve.

    Args:
        interpolation_point_1: The first point at which to interpolate pressure values.
        interpolation_point_2: The second point at which to interpolate pressure values.
        ticks: A sequence of .NET `Tick` values.

    Returns:
        The times at which to set the control points for a leak off curve.
    """
    time_series_interpolation_points = Array.CreateInstance(Double, 2)
    time_series_interpolation_points[0] = interpolation_point_1.Ticks
    time_series_interpolation_points[1] = interpolation_point_2.Ticks
    time_stamp_ticks = Array.CreateInstance(Double, ticks.Length)
    magnitudes = Array.CreateInstance(Double, ticks.Length)
    for i in range(0, ticks.Length):
        tick = ticks[i]
        time_stamp_ticks[i] = tick.Timestamp.Ticks
        magnitudes[i] = tick.Value
    time_series_interpolant = Interpolation.Interpolant1DFactory.CreatePchipInterpolant(time_stamp_ticks,
                                                                                        magnitudes)
    pressure_values = time_series_interpolant.Interpolate(time_series_interpolation_points, 0)
    control_point_times = List[Leakoff.ControlPoint]()
    control_point_times.Add(Leakoff.ControlPoint(
        DateTime=interpolation_point_1,
        Pressure=UnitsNet.Pressure(pressure_values[0], UnitsNet.Units.PressureUnit.PoundForcePerSquareInch)))
    control_point_times.Add(Leakoff.ControlPoint(
        DateTime=interpolation_point_2,
        Pressure=UnitsNet.Pressure(pressure_values[1], UnitsNet.Units.PressureUnit.PoundForcePerSquareInch)))
    return control_point_times


def calculate_leak_off_pressure(leak_off_curve, maximum_pressure_sample):
    """
    Calculate the leak off pressure at the time of maximum pressure.

    Args:
        leak_off_curve: The leak off curve to query.
        maximum_pressure_sample: The sample (magnitude and time) of maximum pressure.

    Returns:

    """
    query_times = List[DateTime]()
    query_times.Add(maximum_pressure_sample.Timestamp)
    leak_off_pressure = leak_off_curve.GetPressureValues(query_times)[0]
    return leak_off_pressure


def calculate_maximum_pressure_sample(stage_part, ticks):
    """
    Calculate the sample (time stamp and magnitude) at which the maximum pressure occurs.

    Args:
        stage_part: The stage part used to limit the queried samples.
        ticks: A iterator of samples for the stage part.

    Returns:
        The sample (time stamp and magnitude) at which the maximum pressure occurs.
    """
    def maximum_pressure_reducer(so_far, candidate):
        if stage_part.StartTime <= candidate.Timestamp <= stage_part.StopTime and candidate.Value > so_far.Value:
            return candidate
        else:
            return so_far

    sentinel_maximum = object_factory.CreateTick[float](DateTime.MinValue, -1000)
    maximum_pressure_sample = functools.reduce(maximum_pressure_reducer, ticks, sentinel_maximum)
    return maximum_pressure_sample


def calculate_stage_part_pressure_samples(native_monitor, stage_part):
    """
    Calculate the pressure samples from the monitor for the `stage_part`.

    Args:
        native_monitor: The .NET `IMonitor` object recording pressures.
        stage_part: The .NET `IStagePart` limiting the monitor times to the stage treatment times.

    Returns:
        The pressure samples from `native_monitor` for the `stage_part`.
    """
    time_range = object_factory.CreateDateTimeOffsetRange(stage_part.StartTime.AddDays(-1),
                                                          stage_part.StopTime.AddDays(1))
    stage_part_pressure_samples = native_monitor.TimeSeries.GetOrderedTimeSeriesHistory(time_range)
    return stage_part_pressure_samples


def calculate_stage_part_visible_time_range(stage_part):
    """
    Calculate the visible time range of the stage treatment.

    Args:
        stage_part: The stage part identifying the stage treatment of interest.

    Returns:
        A `tuple` identifying the start and stop of the stage treatment.
    """
    return stage_part.StartTime.AddHours(-1), stage_part.StopTime.AddHours(1)


def create_leak_off_curve_control_points(leak_off_curve_times):
    """
    Create the control points for an observation.

    Args:
        leak_off_curve_times: The `dict` containing time stamps for specific leak off curve control points.

    Returns:
        The .NET `IList` containing the leak off curve control points.
    """
    leak_off_curve_control_points = List[DateTime]()
    leak_off_curve_control_points.Add(leak_off_curve_times['L1'])
    leak_off_curve_control_points.Add(leak_off_curve_times['L2'])
    return leak_off_curve_control_points


def auto_pick_observation_details(unpicked_observation, native_monitor, stage_part):
    """
    Change `unpicked_observation` by adding details to make it a picked observation.

    Args:
        unpicked_observation: The unpicked observation.
        native_monitor: The .NET `IMonitor` for this observation.
        stage_part: The .NET `IStagePart` observed by `native_monitor`.

    Returns:
        The "picked" observation with the appropriate details filled in.
    """
    # Auto pick observation details to be set
    # - Leak off curve type
    # - Control point times
    # - Visible range x-min time
    # - Visible range x-max time
    # - Position
    # - Delta pressure
    # - Notes
    # - Signal quality

    stage_part_pressure_samples = calculate_stage_part_pressure_samples(native_monitor, stage_part)

    leak_off_curve_times = {
        'L1': stage_part.StartTime.AddMinutes(-20),
        'L2': stage_part.StartTime,
    }
    control_point_times = calculate_leak_off_control_point_times(leak_off_curve_times['L1'],
                                                                 leak_off_curve_times['L2'],
                                                                 stage_part_pressure_samples)

    leak_off_curve = object_factory.CreateLeakoffCurve(Leakoff.LeakoffCurveType.Linear,
                                                       control_point_times)

    maximum_pressure_sample = calculate_maximum_pressure_sample(stage_part, stage_part_pressure_samples)
    leak_off_pressure = calculate_leak_off_pressure(leak_off_curve, maximum_pressure_sample)

    picked_observation = unpicked_observation  # An alias to better communicate intent
    with dnd.disposable(picked_observation.ToMutable()) as mutable_observation :
        mutable_observation.LeakoffCurveType = Leakoff.LeakoffCurveType.Linear
        mutable_observation.ControlPointTimes = create_leak_off_curve_control_points(leak_off_curve_times)
        (mutable_observation.VisibleRangeXminTime,
         mutable_observation.VisibleRangeXmaxTime) = calculate_stage_part_visible_time_range(stage_part)
        mutable_observation.Position = maximum_pressure_sample.Timestamp
        mutable_observation.DeltaPressure = calculate_delta_pressure(leak_off_pressure, maximum_pressure_sample)
        mutable_observation.Notes = "Auto-picked"
        mutable_observation.SignalQuality = Observation.SignalQualityValue.UndrainedCompressive

    return picked_observation


def auto_pick_observations(native_project, native_monitor):
    """
        Automatically pick observations for each treatment stage of `native_project` observed by `native_monitor`.
    Args:
        native_project: The `IProject` whose observations are sought.
        native_monitor: The `IMonitor` whose observations we automatically pick.

    Returns:

    """
    observation_set = object_factory.CreateObservationSet(native_project, 'Auto-picked Observation Set3')

    wells = native_project.Wells.Items
    for well in wells:
        stages = well.Stages.Items
        for stage in stages:

            if is_stage_visible_to_monitor(native_monitor, stage):

                stage_parts = stage.Parts
                for part in stage_parts:

                    # Create unpicked observation
                    unpicked_observation = object_factory.CreateObservation(native_monitor, part)

                    # Auto-pick observation details
                    picked_observation = auto_pick_observation_details(unpicked_observation, native_monitor, part)

                    # Add picked observation to observation set
                    with dnd.disposable(observation_set.ToMutable()) as mutable_observation_set:
                        mutable_observation_set.AddEvent(picked_observation)

    # TODO: Can we use Python disposable decorator?
    # Add observation set to project
    project_with_observation_set = native_project  # An alias to better communicate intent
    with dnd.disposable(native_project.ToMutable()) as mutable_project:
        mutable_project.AddObservationSet(observation_set)

    return project_with_observation_set


def is_stage_visible_to_monitor(native_monitor, stage):
    """
    Determine if the stage treatment is visible to the specified monitor.

    Args:
        native_monitor: The .NET `IMonitor` that may "see" the stage treatment.
        stage: The stage of interest.

    Returns:
        True if the stage is being treated while the monitor is actively monitoring pressures.
    """
    return stage.StartTime.Ticks > native_monitor.StartTime.Ticks and stage.StopTime.Ticks < native_monitor.StopTime.Ticks


def main(cli_args):
    """
    Save project with automatically picked observations from original project read from disk.

    Args:
        cli_args: The command line arguments from `argparse.ArgumentParser`.
    """
    logging.basicConfig(level=logging.INFO)

    # Read Orchid project
    project = orchid.load_project(cli_args.input_project)
    native_project = project.dom_object

    # Automatically pick the observations for a specific monitor
    monitor_name = 'Demo_3H - MonitorWell'
    candidate_monitors = list(project.monitors().find_by_display_name(monitor_name))
    # I actually expect one or more monitors, but I only need one (arbitrarily the first one)
    assert len(candidate_monitors) > 0, (f'One or monitors with display name, "{monitor_name}", expected.'
                                         f' Found {len(candidate_monitors)}.')
    native_monitor = candidate_monitors[0].dom_object
    auto_pick_observations(native_project, native_monitor)

    # Log changed project data if requested
    if cli_args.verbosity >= 2:
        logging.info(f'{native_project.Name=}')
        logging.info(f'{len(native_project.ObservationSets.Items)=}')
        for observation_set in native_project.ObservationSets.Items:
            logging.info(f'{observation_set.Name=}')
            # TODO: Remove when >~ 2021.4
            # logging.info(f'{len(observation_set.LeakOffObservations.Items)=}')
            logging.info(f'{len(observation_set.GetObservations())=}')

    # Save project changes to specified .ifrac file
    orchid.optimized_but_possibly_unsafe_save(project, cli_args.input_project, cli_args.output_project)
    if cli_args.verbosity >= 1:
        logging.info(f'Wrote changes to "{cli_args.output_project}"')


def make_project_path_name(project_dir_name, project_file_name):
    """
    Make a path name to a project.

    Args:
        project_dir_name: The directory name of the project.
        project_file_name: The file name of the project.

    Returns:
        The path name to the .ifrac file for this project.
    """
    return str(project_dir_name.joinpath(project_file_name))


def make_target_file_name_from_source(source_file_name):
    """
    Make a file name for the changed project file name from the original project file name.

    Args:
        source_file_name: The file name of the project originally read.

    Returns:
        The project file name with a `.997` suffix inserted before the `.ifrac` suffix.
    """
    return ''.join([source_file_name.stem, '.997', source_file_name.suffix])


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Automatically pick leak off observations.")
    parser.add_argument('-v', '--verbosity', type=int, choices=[0, 1, 2], default=0,
                        help='Increase output verbosity. (Default: 0; that is, least output.)')

    parser.add_argument('input_project', help=f'Path name of project to read.')

    # Although input file must be 'frankNstein_Bakken_UTM13_FEET.v11.ifrac', I use the original version as the
    # default for the correct output file name.
    default_file_name_to_read = pathlib.Path('frankNstein_Bakken_UTM13_FEET.ifrac')
    default_project_path_name_to_read = make_project_path_name(orchid.training_data_path(),
                                                               default_file_name_to_read)
    default_file_name_to_write = make_target_file_name_from_source(default_file_name_to_read)
    default_project_path_name_to_write = make_project_path_name(orchid.training_data_path(),
                                                                default_file_name_to_write)
    parser.add_argument('-o', '--output_project', default=default_project_path_name_to_write,
                        help=f'Filename of project to write. (Default: {default_project_path_name_to_write}')

    args = parser.parse_args()
    main(args)




#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#


import functools

import pendulum

from orchid import net_date_time as ndt

# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics import IFractureDiagnosticsFactory
# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.Factories import FractureDiagnosticsFactory
# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.Factories.ReferenceCounting import NullReferenceCounterFactory


@functools.lru_cache
def create() -> IFractureDiagnosticsFactory:
    """
    Return an instance of the fracture diagnostics factory used by .NET to construct DOM instances.

    >>> start = pendulum.parse('2022-02-23T15:53:23Z')
    >>> stop = pendulum.parse('2022-02-24T05:54:11Z')
    >>> net_start = ndt.as_net_date_time(start)
    >>> net_stop = ndt.as_net_date_time(stop)
    >>> factory = create()
    >>> date_time_offset_range = factory.CreateDateTimeOffsetRange(net_start, net_stop)
    >>> (date_time_offset_range.Start.ToString('o'), date_time_offset_range.Stop.ToString('o'))
    ('2022-02-23T15:53:23.0000000+00:00', '2022-02-24T05:54:11.0000000+00:00')

    Returns:
        An instance of the fracture diagnostics factory.
    """
    # TODO: We will need a "real" reference counter factory to support deleting DOM instances
    return FractureDiagnosticsFactory.Create(NullReferenceCounterFactory())


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#


import sys

import orchid.configuration

import toolz.curried as toolz

# noinspection PyPackageRequirements
import clr
# noinspection PyUnresolvedReferences
from System import InvalidOperationException


def append_orchid_assemblies_directory_path() -> None:
    """
    Append the directory containing the required Orchid assemblies to `sys.path`.
    """
    orchid_bin_dir = toolz.get_in(['orchid', 'root'], orchid.configuration.python_api())
    if orchid_bin_dir not in sys.path:
        sys.path.append(orchid_bin_dir)


append_orchid_assemblies_directory_path()

clr.AddReference('Orchid.FractureDiagnostics.SDKFacade')
# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.SDKFacade import ScriptAdapter


class ScriptAdapterContext:
    """
    A "private" class with the responsibility to initialize and shutdown the .NET ScriptAdapter class.

    I considered making `ProjectStore` a context manager; however, the API then becomes somewhat unclear.

        - Does the constructor enter the context? Must a caller initialize the instance and then enter the
          context?
        - What results if a caller *does not* enter the context?
        - Enters the context twice?

    Because I was uncertain I created this private class to model the `ScriptAdapter` context. The property,
    `ProjectStore.native_project`, enters the context if it will actually read the project and exits the
    context when the read operation is finished.

    For information on Python context managers, see
    [the Python docs](https://docs.python.org/3.8/library/stdtypes.html#context-manager-types)
    """

    def __enter__(self):
        try:
            ScriptAdapter.Init()
            return self
        # TODO: Correct exception type / DEADFALL issue
        except InvalidOperationException as ioe:
            if 'REVEAL-CORE-0xDEADFA11' in ioe.Message:
                print('Orchid licensing error. Please contact Orchid technical support.')
                sys.exit(-1)
            else:
                raise

    def __exit__(self, exc_type, exc_val, exc_tb):
        ScriptAdapter.Shutdown()
        # Returning no value will propagate the exception to the caller in the normal way
        return



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#


"""
Example of adding stages to a project using the low-level (Python.NET) API.
"""


import argparse
from collections import namedtuple
import dataclasses as dc
import logging
import pathlib
from typing import Optional

import pendulum  # Used for creating time-zone aware date times (UTC by default)

import orchid
from orchid import (
    dot_net_disposable as dnd,
    native_stage_adapter as nsa,
    native_well_adapter as nwa,
    net_fracture_diagnostics_factory as net_factory,
    net_quantity as onq,
    net_date_time as ndt,
    measurement as om,
    unit_system as units,
)

import clr
clr.AddReference('System.Collections')

# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics import (IStage, IStagePart)
# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.SDKFacade import ScriptAdapter
# noinspection PyUnresolvedReferences
from System import (Array, UInt32, Nullable,)
# noinspection PyUnresolvedReferences
from System.Collections.Generic import List
# noinspection PyUnresolvedReferences
from UnitsNet import Pressure


object_factory = net_factory.create()


@dc.dataclass
class CreateStageDto:
    order_of_completion_on_well: int  # Must be non-negative
    connection_type: nsa.ConnectionType
    md_top: om.Quantity  # Must be length
    md_bottom: om.Quantity  # Must be length
    maybe_shmin: Optional[om.Quantity] = None  # If not None, must be pressure
    cluster_count: int = 0  # Must be non-negative
    # WARNING: one need supply neither a start time nor a stop time; however, not supplying this data can
    # produce unexpected behavior for the `global_stage_sequence_number` property. For example, one can
    # generate duplicate values for the `global_stage_sequence_number`. This unexpected behavior is a known
    # issue with Orchid.
    #
    # Note supplying no value (an implicit `None`) results in the largest possible .NET time range.
    maybe_time_range: Optional[pendulum.Period] = None

    # WARNING: one must currently supply an ISIP for each stage; otherwise, Orchid fails to correctly load
    # the project saved with the added stages.
    maybe_isip: Optional[om.Quantity] = None  # The actual value must be a pressure

    def create_stage(self, well: nwa.NativeWellAdapter):
        # Must supply the unit system for conversions
        native_md_top = onq.as_net_quantity(units.UsOilfield.LENGTH, self.md_top)
        native_md_bottom = onq.as_net_quantity(units.UsOilfield.LENGTH, self.md_bottom)
        native_shmin = (ScriptAdapter.MakeOptionSome(onq.as_net_quantity(units.UsOilfield.PRESSURE,
                                                                         self.maybe_shmin))
                        if self.maybe_shmin is not None
                        else ScriptAdapter.MakeOptionNone[Pressure]())
        no_time_range_native_stage = object_factory.CreateStage(
            UInt32(self.order_of_completion_on_well),
            well.dom_object,
            self.connection_type.value,
            native_md_top,
            native_md_bottom,
            native_shmin,
            UInt32(self.cluster_count)
        )
        with dnd.disposable(no_time_range_native_stage.ToMutable()) as mutable_stage:
            if self.maybe_time_range is not None:
                stage_part = object_factory.CreateStagePart(no_time_range_native_stage,
                                                            ndt.as_net_date_time(self.maybe_time_range.start),
                                                            ndt.as_net_date_time(self.maybe_time_range.end),
                                                            onq.as_net_quantity_in_specified_unit(
                                                                units.UsOilfield.PRESSURE, self.maybe_isip))
                stage_parts = List[IStagePart]()
                stage_parts.Add(stage_part)
                mutable_stage.Parts = stage_parts
        native_stage = no_time_range_native_stage

        return nsa.NativeStageAdapter(native_stage)

    def __post_init__(self):
        # See the
        # [StackOverflow post](https://stackoverflow.com/questions/54488765/validating-input-when-mutating-a-dataclass)
        assert self.order_of_completion_on_well >= 0, f'order_of_completion_on_well must be non-zero'
        assert self.md_top.check('[length]'), f'md_top must be a length'
        assert self.md_bottom.check('[length]'), f'md_bottom must be a length'
        if self.maybe_shmin is not None:
            assert self.maybe_shmin.check('[pressure]'), f'maybe_shmin must be a pressure if not `None`'
        assert self.cluster_count >= 0, f'cluster_count must be non-zero'
        if self.maybe_isip is not None:
            assert self.maybe_isip.check('[pressure]'), f'maybe_isip must be a pressure if not `None`'


CreatedStageDetails = namedtuple('CreatedStageDetails', ['name', 'shmin', 'cluster_count',
                                                         'global_stage_sequence_no', 'start_time', 'stop_time'])


def add_stages(project):
    # Find well to which to add stages
    candidate_well_name = 'Demo_4H'
    candidate_wells = list(project.wells().find_by_display_name(candidate_well_name))
    assert len(candidate_wells) == 1, (f'Expected single well named "{candidate_well_name}".'
                                       f' Found {len(candidate_wells)}.')
    target_well = candidate_wells[0]

    # Create an iterable of stages to append
    stages_to_append = [
        CreateStageDto(
            35,  # hard-coded to be one greater than largest `order_of_completion_on_well`
            nsa.ConnectionType.PLUG_AND_PERF,
            orchid.make_measurement(orchid.unit_system.UsOilfield.LENGTH, 20898.2),
            orchid.make_measurement(orchid.unit_system.UsOilfield.LENGTH, 20985.8),
            # `pendulum` uses UTC by default for timezone (and UTC required)
            maybe_time_range=pendulum.parse('2018-06-06T05:34:03.6839387/2018-06-06T07:19:35.5601864'),
            maybe_isip=3420.32 * orchid.unit_registry.psi,
        ),
        CreateStageDto(
            36,
            nsa.ConnectionType.PLUG_AND_PERF,
            orchid.make_measurement(orchid.unit_system.UsOilfield.LENGTH, 17362.2),
            orchid.make_measurement(orchid.unit_system.UsOilfield.LENGTH, 17372.3),
            maybe_shmin=orchid.make_measurement(orchid.unit_system.UsOilfield.PRESSURE, 2.322),
            # `pendulum` uses UTC by default for timezone (and UTC required)
            maybe_time_range=pendulum.parse('2018-06-15T14:11:40.450044/2018-06-15T15:10:11.200044'),
            maybe_isip=2712.70 * orchid.unit_registry.psi,
        ),
        CreateStageDto(
            37,
            nsa.ConnectionType.PLUG_AND_PERF,
            orchid.make_measurement(orchid.unit_system.UsOilfield.LENGTH, 10627.2),
            orchid.make_measurement(orchid.unit_system.UsOilfield.LENGTH, 10759.7),
            cluster_count=7,
            # `pendulum` uses UTC by default for timezone (and UTC required)
            maybe_time_range=pendulum.parse('2018-06-28T23:35:54.3790545/2018-06-29T01:18:05.8397489'),
            maybe_isip=3192.69 * orchid.unit_registry.psi
        ),
    ]
    created_stages = [stage_dto.create_stage(target_well) for stage_dto in stages_to_append]

    for created_stage in created_stages:
        logging.info(CreatedStageDetails(created_stage.name,
                                         f'{created_stage.shmin:.3f~P}' if created_stage.shmin else 'None',
                                         created_stage.cluster_count,
                                         created_stage.global_stage_sequence_number,
                                         str(created_stage.start_time),
                                         str(created_stage.stop_time)))

    # Add stages to target_well
    with dnd.disposable(target_well.dom_object.ToMutable()) as mutable_well:
        native_created_stages = Array[IStage]([created_stage.dom_object for created_stage in created_stages])
        mutable_well.AddStages(native_created_stages)


def main(cli_args):
    """
    Add stages to an existing project and save changes back to disk.

    Args:
        cli_args: The command line arguments from `argparse.ArgumentParser`.
    """
    logging.basicConfig(level=logging.INFO)

    # Read Orchid project
    project = orchid.load_project(cli_args.input_project)

    add_stages(project)

    # Save project changes to specified .ifrac file
    orchid.optimized_but_possibly_unsafe_save(project, cli_args.input_project, cli_args.output_project)
    if cli_args.verbosity >= 1:
        logging.info(f'Wrote changes to "{cli_args.output_project}"')


def make_project_path_name(project_dir_name, project_file_name):
    """
    Make a path name to a project.

    Args:
        project_dir_name: The directory name of the project.
        project_file_name: The file name of the project.

    Returns:
        The path name to the .ifrac file for this project.
    """
    return str(project_dir_name.joinpath(project_file_name))


def make_target_file_name_from_source(source_file_name):
    """
    Make a file name for the changed project file name from the original project file name.

    Args:
        source_file_name: The file name of the project originally read.

    Returns:
        The project file name with a `.996` suffix inserted before the `.ifrac` suffix.
    """
    return ''.join([source_file_name.stem, '.996', source_file_name.suffix])


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Adds newly created stages to a well in a project")
    parser.add_argument('-v', '--verbosity', type=int, choices=[0, 1, 2], default=0,
                        help='Increase output verbosity. (Default: 0; that is, least output.)')

    parser.add_argument('input_project', help=f'Path name of project to read.')

    # Although input file must be 'frankNstein_Bakken_UTM13_FEET.v11.ifrac', I use the original version as the
    # default for the correct output file name.
    default_file_name_to_read = pathlib.Path('frankNstein_Bakken_UTM13_FEET.ifrac')
    default_project_path_name_to_read = make_project_path_name(orchid.training_data_path(),
                                                               default_file_name_to_read)
    default_file_name_to_write = make_target_file_name_from_source(default_file_name_to_read)
    default_project_path_name_to_write = make_project_path_name(orchid.training_data_path(),
                                                                default_file_name_to_write)
    parser.add_argument('-o', '--output_project', default=default_project_path_name_to_write,
                        help=f'Filename of project to write. (Default: {default_project_path_name_to_write}')

    args = parser.parse_args()
    main(args)




#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#

"""
This notebook illustrates using the Orchid* Python API and the pandas package to
perform VFR analysis

(*Orchid is a mark of Reveal Energy Services, Inc)
"""

# Example: Using Pandas to Analyze Volume to First Response (VFR)

# 0.5 Import packages

# The only import needed for the Python API is `orchid` itself.

import orchid

# The remaining imports are standard python packages to support the analysis.

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import scipy.signal as signal
import datetime as dt


# 1.0 Defining Support Functions

# Convenient constants

FIRST_DERIVATIVE_TIME = 'First Derivative Time'
SECOND_DERIVATIVE_TIME = 'Second Derivative Time'
V2FR_FIRST_DERIVATIVE = 'VFR (First Derivative)'
V2FR_SECOND_DERIVATIVE = 'VFR (Second Derivative)'
FIRST_DERIVATIVE_EVENT_TIME = 'First Derivative Event Time'
SECOND_DERIVATIVE_EVENT_TIME = 'Second Derivative Event Time'

# Completion Support Function

# This function will support taking in a stop time and calculating the treatment aggregates
def compute_stage_treatment_aggregates(treatment_stage, stop_time):
    # These calculations IGNORE all calculation warnings.
    pumped_fluid_volume = orchid.pumped_fluid_volume(treatment_stage,
                                                     treatment_stage.start_time,
                                                     stop_time)[0].magnitude
    total_proppant_mass = orchid.total_proppant_mass(treatment_stage,
                                                     treatment_stage.start_time,
                                                     stop_time)[0].magnitude
    median_treating_pressure = orchid.median_treating_pressure(treatment_stage,
                                                               treatment_stage.start_time,
                                                               stop_time)[0].magnitude
    return pumped_fluid_volume, total_proppant_mass, median_treating_pressure


# The first event-detection algorithm is to find when the first derivative exceeds a threshold value
def first_derivative_threshold(pressure_curve, threshold, window=51, poly=3):
    times = pressure_curve.index.values
    average_dt = np.mean(np.diff(pressure_curve.index)).total_seconds() / 60
    pressure = pressure_curve.values
    first_derivative = signal.savgol_filter(pressure,
                                            window_length=window,
                                            polyorder=poly,
                                            delta=average_dt,
                                            deriv=1)

    ndx = np.argwhere(first_derivative > threshold)

    return times[ndx[0][0]] if len(ndx) > 0 else None


# The second event-detection algorithm uses the second derivative which is fed into a peak finding algorithm
def second_derivative_peak(pressure_curve, window=51, poly=3):
    times = pressure_curve.index.values
    average_dt = np.mean(np.diff(pressure_curve.index)).total_seconds() / 60
    pressure = pressure_curve.values
    second_derivative = signal.savgol_filter(pressure,
                                             window_length=window,
                                             polyorder=poly,
                                             delta=average_dt,
                                             deriv=2)

    peaks = signal.find_peaks(second_derivative, height=0.2, distance=30, width=5)
    return times[peaks[0][0]] if len(peaks[0]) > 0 else None


def data_points_from_time_series_in_target_units(pressure_series, target_units):
    # ts is short variable name for data_points data
    ts_units = pressure_series.sampled_quantity_unit()
    ts = pressure_series.data_points()
    # Convert list of magnitudes (no units) to `numpy` array including units for faster operations
    ts_values_w_units = orchid.unit_registry.Quantity(ts.to_numpy(), ts_units.value[0])
    ts_in_target_units = ts_values_w_units.to(target_units.value[0])
    return pd.Series(data=ts_in_target_units.magnitude, index=ts.index)


def calculate_volume_2_first_response():
    # 2.0 Load the .ifrac project

    # The following code simply captures the configured location of the Orchid training data. It is not needed to
    # use the Orchid Python API itself, but it is used in this example to load well-known data.
    orchid_training_data_path = orchid.training_data_path()

    print('Wait patiently for the project to load...')
    permian_project = orchid.load_project(str(orchid_training_data_path.joinpath(
        'Project_frankNstein_Permian_UTM13_FEET.ifrac')))

    # 3.0 Extract Data and Apply Event Detectors

    # Takes a few minutes to run
    print('...And even more patiently for the calculations to complete...')
    p_time_series = list(permian_project.time_series().find_by_display_name('P1-Downhole-12550-Pressure'))
    p_time_series = p_time_series[0]  # Simply take the first matching time series
    p_data = data_points_from_time_series_in_target_units(p_time_series, orchid.unit_system.UsOilfield.PRESSURE)

    vfr_data = []
    for well in permian_project.wells().all_objects():
        for stage in well.stages().all_objects():
            stage_start_time = stage.start_time
            stage_stop_time = stage.stop_time

            # p_stg is the monitor pressure data with the given stage start/stop time
            p_stg = p_data[stage_start_time:stage_stop_time]
            derive_1_time = first_derivative_threshold(p_stg, 0.2)
            derive_2_time = second_derivative_peak(p_stg)
            if derive_1_time is not None:
                derive_1_time = dt.datetime.utcfromtimestamp(derive_1_time.tolist() * 1e-9).replace(tzinfo=orchid.UTC)
            if derive_2_time is not None:
                derive_2_time = dt.datetime.utcfromtimestamp(derive_2_time.tolist() * 1e-9).replace(tzinfo=orchid.UTC)
            vfr_d1, _, _ = (compute_stage_treatment_aggregates(stage, derive_1_time) if derive_1_time is not None
                            else (None, None, None))
            vfr_d2, _, _ = (compute_stage_treatment_aggregates(stage, derive_2_time) if derive_2_time is not None
                            else (None, None, None))
            vfr_data.append((well.name, stage.display_stage_number, derive_1_time, vfr_d1, derive_2_time, vfr_d2))

    # Show the dataframe
    df = pd.DataFrame(data=vfr_data, columns=['Well', 'Stage', FIRST_DERIVATIVE_TIME, V2FR_FIRST_DERIVATIVE,
                                              SECOND_DERIVATIVE_TIME, V2FR_SECOND_DERIVATIVE])
    print()
    print('Volume 2 First Response')
    print(df.to_string(max_rows=10))
    input('Press enter to continue...')

    # 3.1 Visualize the Pressure Data with Event Locations

    # Example with a selected well and stage number

    # Find a particularly interesting well and stage to visualize
    candidate_wells = list(permian_project.wells().find_by_display_name('C1'))
    well_0 = candidate_wells[0]
    selected_stage = well_0.stages().find_by_display_stage_number(8)

    p_stg = p_data[selected_stage.start_time:selected_stage.stop_time]
    plt.plot(p_stg, label='PData')
    ymin, ymax = plt.ylim()

    selected_ndx = 7  # Matches the DF index
    d1_x = df.iloc[selected_ndx][FIRST_DERIVATIVE_TIME]
    if d1_x is not None:
        plt.vlines(d1_x, ymin, ymax,  'r', label=FIRST_DERIVATIVE_EVENT_TIME)

    d2_x = df.iloc[selected_ndx][SECOND_DERIVATIVE_TIME]
    if d2_x is not None:
        plt.vlines(d2_x, ymin, ymax,  'k', label=SECOND_DERIVATIVE_EVENT_TIME)

    plt.legend()
    plt.xlabel('Time')
    plt.ylabel('Pressure')
    plt.title('VFR Analysis with Multiple Event Detectors')

    plt.show()


def main():
    print(__doc__)
    calculate_volume_2_first_response()


if __name__ == '__main__':
    main()



#  Copyright (c) 2017-2022 Reveal Energy Services, Inc 
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#

import glob
import logging
import os
import pathlib
from typing import Dict
import warnings

import toolz.curried as toolz
import yaml

import orchid.version


_logger = logging.getLogger(__name__)


class ConfigurationError(Exception):
    pass


# Constants for environment variable names
ORCHID_ROOT_ENV_VAR = 'ORCHID_ROOT'
ORCHID_TRAINING_DATA_ENV_VAR = 'ORCHID_TRAINING_DATA'


def get_environment_configuration() -> Dict:
    """
    Gets the API configuration from the system environment.

    Returns:
        The configuration, if any, calculated from the system environment.
    """
    environment_configuration = {}
    if ORCHID_ROOT_ENV_VAR in os.environ and ORCHID_TRAINING_DATA_ENV_VAR in os.environ:
        environment_configuration = {'orchid': {'root': os.environ[ORCHID_ROOT_ENV_VAR],
                                                'training_data': os.environ[ORCHID_TRAINING_DATA_ENV_VAR]}}
    elif ORCHID_ROOT_ENV_VAR in os.environ:
        environment_configuration = {'orchid': {'root': os.environ[ORCHID_ROOT_ENV_VAR]}}
    elif ORCHID_TRAINING_DATA_ENV_VAR in os.environ:
        environment_configuration = {'orchid': {'training_data': os.environ[ORCHID_TRAINING_DATA_ENV_VAR]}}

    _logger.debug(f'environment configuration = {environment_configuration}')

    return environment_configuration


def get_fallback_configuration() -> Dict:
    """
    Returns final fallback API configuration.

    Returns:
        A Python dictionary with the default (always available configuration).

    Warning:
        Although we have striven to make the default configuration a working configuration, we can only ensure
        that the default configuration meets the minimal "syntax" required by the Python API. For example, if
        Orchid is **not** installed in the default location, and the `directory` key is not overridden by a
        higher priority configuration, the Python API will **fail** to load the Orchid assemblies and throw
        an exception at runtime.
    """

    # Symbolically, the standard location for the installed Orchid binaries is
    # `$ProgramFiles/Reveal Energy Services, Inc/Orchid/<version-specific-directory>`. The following code
    # calculates an actual location by substituting the current version number for the symbol,
    # `<version-specific-directory>`.
    standard_orchid_dir = pathlib.Path(os.environ['ProgramFiles']).joinpath('Reveal Energy Services',
                                                                            'Orchid')
    version_id = orchid.version.api_version()
    version_dirname = f'Orchid-{version_id.major}.{version_id.minor}.{version_id.micro}'
    glob_path = standard_orchid_dir.joinpath(f'{version_dirname}*', 'PythonApiLibs')
    candidate_directories = glob.glob(str(glob_path))
    if len(candidate_directories) == 1:
        fallback = {'orchid': {'root': str(candidate_directories[0])}}
    elif len(candidate_directories) == 0:
        fallback = {}
    else:
        fallback = {}
        warnings.warn(f'Fallback configuration found multiple matches for {str(glob_path)}')
    _logger.debug(f'fallback configuration={fallback}')
    return fallback


def get_file_configuration() -> Dict:
    """
    Returns the API configuration read from the file system.

    Returns:
        A python dictionary with the default (always available configuration).
    """

    # This code looks for the configuration file, `python_api.yaml`, in the `.orchid` sub-directory of the
    # user-specific (and system-specific) home directory. See the Python documentation of `home()` for
    # details.
    file = {}
    file_config_path = pathlib.Path.home().joinpath('.orchid', 'python_api.yaml')
    if file_config_path.exists():
        with file_config_path.open('r') as in_stream:
            file = yaml.full_load(in_stream)
    _logger.debug(f'file configuration={file}')
    return file


def python_api() -> Dict[str, str]:
    """
    Calculate the configuration for the Python API.

        Returns: The Python API configuration.
    """

    fallback_configuration = get_fallback_configuration()
    file_configuration = get_file_configuration()
    env_configuration = get_environment_configuration()

    result = merge_configurations(fallback_configuration, file_configuration, env_configuration)
    _logger.debug(f'result configuration={result}')
    return result


def merge_configurations(fallback_configuration, file_configuration, env_configuration):
    # The rules for merging these configurations is not the same as a simple dictionary. The rules are:
    # - If two different configurations share a top-level key, merge the second level dictionaries.
    # - Then merge the distinct top-level keys.
    distinct_top_level_keys = set(toolz.concat([fallback_configuration.keys(),
                                                file_configuration.keys(),
                                                env_configuration.keys()]))
    result = {}
    for top_level_key in distinct_top_level_keys:
        fallback_child_configuration = fallback_configuration.get(top_level_key, {})
        file_child_configuration = file_configuration.get(top_level_key, {})
        env_child_configuration = env_configuration.get(top_level_key, {})
        child_configuration = toolz.merge(fallback_child_configuration,
                                          file_child_configuration,
                                          env_child_configuration)
        result[top_level_key] = child_configuration
    return result


def training_data_path() -> pathlib.Path:
    """
    Returns the path of the directory containing the Orchid training data.

    Returns:
        The Orchid training data path.

    Raises:
        This function raises KeyError if the training directory path is not available from the package
        configuration.
    """
    result = pathlib.Path(toolz.get_in(['orchid', 'training_data'], python_api()))
    return result



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#


"""
Low-level example of creating multi-picking events using the low-level API.
"""

import argparse
import logging
import pathlib

import clr
import orchid
from orchid import (
    dot_net_disposable as dnd,
    net_fracture_diagnostics_factory as net_factory,
)

# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics import (MonitorExtensions, Leakoff, Observation)
# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.Factories.Implementations import LeakoffCurves, MultiPickingObservation
# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.SDKFacade import (
    ScriptAdapter,
)
# noinspection PyUnresolvedReferences
from System import (Array, Double, DateTime, String)
# noinspection PyUnresolvedReferences
from System.IO import (FileStream, FileMode, FileAccess, FileShare)
# noinspection PyUnresolvedReferences
import UnitsNet

clr.AddReference('Orchid.Math')
clr.AddReference('System.Collections')
# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics import IStagePart
# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.MultiPickingObservation import Classification as MultiPickingObservationClassification
# noinspection PyUnresolvedReferences
from Orchid.Math import Interpolation
# noinspection PyUnresolvedReferences
from System import DateTime
# noinspection PyUnresolvedReferences
from System.Collections.Generic import List


object_factory = net_factory.create()


def multi_pick_observations(native_project, native_monitor):
    """
        Multi-pick observations for each treatment stage of `native_project` observed by `native_monitor`.
    Args:
        native_project: The `IProject` whose observations are sought.
        native_monitor: The `IMonitor` whose observations we automatically pick.

    Returns:

    """
    observation_set = object_factory.CreateObservationSet(native_project, 'Multi-pick Observation Set')

    # Arbitrarily select well, 'Demo_1H'
    monitored_well = [well for well in native_project.Wells.Items if well.Name == 'Demo_1H'][0]
    # Arbitrarily select stages and classifications. All stages are visible with a single stage part.
    monitored_stage_numbers = {
        37: MultiPickingObservationClassification.Spacer,
        38: MultiPickingObservationClassification.DiverterDrop,
        44: MultiPickingObservationClassification.Chemical,
        16: MultiPickingObservationClassification.RateChange,
        33: MultiPickingObservationClassification.BadData,
    }
    monitored_stages = [stage for stage in monitored_well.Stages.Items if
                        stage.DisplayStageNumber in monitored_stage_numbers]
    for stage in monitored_stages:
        if is_stage_visible_to_monitor(native_monitor, stage):
            stage_parts = stage.Parts
            for part in stage_parts:
                # Create multi-pick observation
                picked_observation = object_factory.CreateMultiPickingEventObservation(
                    part, part.StartTime, monitored_stage_numbers[stage.DisplayStageNumber],
                    f'auto-pick: {monitored_stage_numbers[stage.DisplayStageNumber]}')

                # One must supply a value for `picked_observation.Name`. If no value is supplied, no error occurs when:
                # - Creating the observation
                # - Adding the observation to the observation set
                # - Adding the observation set to the project
                # - Saving the project.
                #
                # However, if one attempts to load such a project into Orchid, Orchid will report a "corrupted
                # project" because at "object name is empty".
                with dnd.disposable(picked_observation.ToMutable()) as mutable_observation:
                    # The function `MultiPickingObservation` generates a name using the second argument as a "template".
                    # The returned name is guaranteed to be unique within the specified observation set. One can
                    # actually use any name that is unique within the observation set and is meaningful to engineers.
                    unique_observation_name = MultiPickingObservation.GetUniqueName(observation_set,
                                                                                    native_monitor.Name)
                    mutable_observation.Name = unique_observation_name

                # Add picked observation to observation set
                with dnd.disposable(observation_set.ToMutable()) as mutable_observation_set:
                    mutable_observation_set.AddEvent(picked_observation)

    # Add observation set to project
    project_with_observation_set = native_project  # An alias to better communicate intent
    with dnd.disposable(native_project.ToMutable()) as mutable_project:
        mutable_project.AddObservationSet(observation_set)

    return project_with_observation_set


# TODO: Move this utility function to the high-level API
# Since I copied the skeleton of this code from `auto_pick.py`, this function is probably copied in
# - `auto_pick.py`
# - `auto_pick_and_create_attributes.py`
# - `auto_pick_iterate_example.py`
def is_stage_visible_to_monitor(native_monitor, stage):
    """
    Determine if the stage treatment is visible to the specified monitor.

    Args:
        native_monitor: The .NET `IMonitor` that may "see" the stage treatment.
        stage: The stage of interest.

    Returns:
        True if the stage is being treated while the monitor is actively monitoring pressures.
    """
    return (stage.StartTime.Ticks > native_monitor.StartTime.Ticks and
            stage.StopTime.Ticks < native_monitor.StopTime.Ticks)


def main(cli_args):
    """
    Save project with automatically picked observations from original project read from disk.

    Args:
        cli_args: The command line arguments from `argparse.ArgumentParser`.
    """
    logging.basicConfig(level=logging.INFO)

    # Read Orchid project
    project = orchid.load_project(cli_args.input_project)
    native_project = project.dom_object

    # Automatically pick the observations for a specific monitor
    monitor_name = 'Demo_3H - MonitorWell'
    candidate_monitors = list(project.monitors().find_by_display_name(monitor_name))
    # I actually expect one or more monitors, but I only need one (arbitrarily the first one)
    assert len(candidate_monitors) > 0, (f'One or monitors with display name, "{monitor_name}", expected.'
                                         f' Found {len(candidate_monitors)}.')
    native_monitor = candidate_monitors[0].dom_object
    multi_pick_observations(native_project, native_monitor)

    # Log changed project data if requested
    if cli_args.verbosity >= 2:
        logging.info(f'{native_project.Name=}')
        logging.info(f'{len(native_project.ObservationSets.Items)=}')
        for observation_set in native_project.ObservationSets.Items:
            logging.info(f'{observation_set.Name=}')
            logging.info(f'{len(observation_set.LeakOffObservations.Items)=}')
            logging.info(f'{len(observation_set.MultiPickingObservations.Items)=}')

    # Save project changes to specified .ifrac file
    orchid.optimized_but_possibly_unsafe_save(project, cli_args.input_project, cli_args.output_project)
    if cli_args.verbosity >= 1:
        logging.info(f'Wrote changes to "{cli_args.output_project}"')


def make_project_path_name(project_dir_name, project_file_name):
    """
    Make a path name to a project.

    Args:
        project_dir_name: The directory name of the project.
        project_file_name: The file name of the project.

    Returns:
        The path name to the .ifrac file for this project.
    """
    return str(project_dir_name.joinpath(project_file_name))


def make_target_file_name_from_source(source_file_name):
    """
    Make a file name for the changed project file name from the original project file name.

    Args:
        source_file_name: The file name of the project originally read.

    Returns:
        The project file name with a `.992` suffix inserted before the `.ifrac` suffix.
    """
    return ''.join([source_file_name.stem, '.992', source_file_name.suffix])


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Create multi-picking events.")
    parser.add_argument('-v', '--verbosity', type=int, choices=[0, 1, 2], default=0,
                        help='Increase output verbosity. (Default: 0; that is, least output.)')

    parser.add_argument('input_project', help=f'Path name of project to read.')

    # Although input file must be 'frankNstein_Bakken_UTM13_FEET.v11.ifrac', I use the original version as the
    # default for the correct output file name.
    default_file_name_to_read = pathlib.Path('frankNstein_Bakken_UTM13_FEET.ifrac')
    default_project_path_name_to_read = make_project_path_name(orchid.training_data_path(),
                                                               default_file_name_to_read)
    default_file_name_to_write = make_target_file_name_from_source(default_file_name_to_read)
    default_project_path_name_to_write = make_project_path_name(orchid.training_data_path(),
                                                                default_file_name_to_write)
    parser.add_argument('-o', '--output_project', default=default_project_path_name_to_write,
                        help=f'Filename of project to write. (Default: {default_project_path_name_to_write}')

    args = parser.parse_args()
    main(args)




#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#


import logging


logging.warning('This package is a placeholder.')
logging.warning('Import `orchid` to actually use the Orchid* Python API.')
logging.warning('(*Orchid is a mark of Reveal Energy Services. Inc.)')



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#
from toolz import curried as toolz


def constantly(x):
    """
    Creates a function that always returns the value, `x`, independent of **all** arguments passed to it..
    Args:
        x: The value to return

    Returns:
        Returns a function takes any arguments yet always returns `x`.
    """

    # noinspection PyUnusedLocal
    def make_constantly(*args, **kwargs):
        return toolz.identity(x)

    return make_constantly



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#

"""
This script illustrates using the Orchid* Python API and the `matplotlib` package to plot well trajectories.

(*Orchid is a mark of Reveal Energy Services, Inc)
"""


# Example: Plotting well trajectories

# 0.5 Import packages

# The only import needed for the Orchid Python API is `orchid` itself.
import orchid

# The remaining imports are standard python packages to support the analysis.

import matplotlib.pyplot as plt
# The following import is included for its "side-effects" of an improved color schemes and
# plot styles. (See the "Tip" in section 9.2 of "Python for Data Analysis" for details.)
import seaborn as sns


def plot_trajectories():
    """Plot the trajectories for a single well."""

    # The following code simply captures the configured location of the Orchid training data. It is not needed to
    # use the Orchid Python API itself, but it is used in this example to load well-known data.
    orchid_training_data_path = orchid.training_data_path()

    # 1.0 Load the .ifrac project

    project = orchid.load_project(str(orchid_training_data_path.joinpath(
        'Project-frankNstein_Montney_UTM13_METERS.ifrac')))

    # 1.1 Get the trajectories for each well of the project

    all_wells = project.wells().all_objects()
    wells = dict([(w.name, w) for w in all_wells])
    trajectories = {wn: w.trajectory for (wn, w) in wells.items()}

    # 1.2 Get the eastings ond northings in the project reference frame

    eastings = {wn: t.get_easting_array(orchid.WellReferenceFrameXy.PROJECT) for (wn, t) in trajectories.items()}
    northings = {wn: t.get_northing_array(orchid.WellReferenceFrameXy.PROJECT) for (wn, t) in trajectories.items()}

    # 2.0 Plot the eastings and northings for each trajectory

    default_well_colors = project.default_well_colors()
    for (i, wn) in enumerate(wells.keys()):
        plt.plot(eastings[wn], northings[wn],
                 label=f'{wells[wn].display_name}',
                 color=default_well_colors[i % len(default_well_colors)])
    plt.title(f'{project.name} Well Trajectories (Project Coordinates)')
    plt.legend(loc='best')
    plt.xlabel(f'Easting ({orchid.abbreviation(project.project_units.LENGTH)})')
    plt.ylabel(f'Northing ({orchid.abbreviation(project.project_units.LENGTH)})')

    plt.rcParams['figure.dpi'] = 150
    plt.show()


def main():
    print(__doc__)
    plot_trajectories()


if __name__ == '__main__':
    main()



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#

# We must use IntEnum for accurate comparison in testing because PythonNet converts .NET enum values to integers.
from enum import IntEnum

# noinspection PyUnresolvedReferences,PyPackageRequirements
from Orchid.FractureDiagnostics import WellReferenceFrameXy as NetWellReferenceFrameXy
# noinspection PyUnresolvedReferences,PyPackageRequirements
from Orchid.FractureDiagnostics import DepthDatum as NetDepthDatum


class WellReferenceFrameXy(IntEnum):
    ABSOLUTE_STATE_PLANE = NetWellReferenceFrameXy.AbsoluteStatePlane
    PROJECT = NetWellReferenceFrameXy.Project
    WELL_HEAD = NetWellReferenceFrameXy.WellHead


class DepthDatum(IntEnum):
    GROUND_LEVEL = NetDepthDatum.GroundLevel
    KELLY_BUSHING = NetDepthDatum.KellyBushing
    SEA_LEVEL = NetDepthDatum.SeaLevel



#  Copyright (c) 2017-2022 Reveal Energy Services, Inc 
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#

from collections import namedtuple
from typing import Iterable

import option
import toolz.curried as toolz

import orchid.base
from orchid import (
    dot_net_dom_access as dna,
    dot_net_disposable as dnd,
    dom_project_object as dpo,
    searchable_stages as oss,
    measurement as om,
    native_stage_adapter as nsa,
    native_subsurface_point as nsp,
    native_trajectory_adapter as nta,
    net_quantity as onq,
    reference_origins as origins,
)

# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics import IStage, IWell
# noinspection PyUnresolvedReferences
import UnitsNet
# noinspection PyUnresolvedReferences
from System import Array, UInt32

WellHeadLocation = namedtuple('WellHeadLocation',
                              ['easting', 'northing', 'depth'])


def replace_no_uwi_with_text(uwi):
    return uwi if uwi else 'No UWI'


class NativeWellAdapter(dpo.DomProjectObject):
    """Adapts a native IWell to python."""

    def __init__(self, net_well: IWell):
        """
        Constructs an instance adapting a .NET IWell.

        Args:
            net_well: The .NET well to be adapted.
        """
        super().__init__(net_well, orchid.base.constantly(net_well.Project))

    trajectory = dna.transformed_dom_property('trajectory', 'The trajectory of the adapted .NET well.',
                                              nta.NativeTrajectoryAdapterIdentified)
    uwi = dna.transformed_dom_property('uwi', 'The UWI of the adapted .', replace_no_uwi_with_text)

    # The formation property **does not** check when a `None` value is passed from Orchid.
    # Although it is possible, it is very unlikely to occur from IWell.Formation.
    formation = dna.dom_property('formation', 'The production formation the well is landed')

    @property
    def ground_level_elevation_above_sea_level(self) -> om.Quantity:
        return onq.as_measurement(self.expect_project_units.LENGTH,
                                  option.maybe(self.dom_object.GroundLevelElevationAboveSeaLevel))

    @property
    def kelly_bushing_height_above_ground_level(self) -> om.Quantity:
        return onq.as_measurement(self.expect_project_units.LENGTH,
                                  option.maybe(self.dom_object.KellyBushingHeightAboveGroundLevel))

    @property
    def wellhead_location(self):
        dom_whl = self.dom_object.WellHeadLocation
        result = toolz.pipe(dom_whl,
                            toolz.map(option.maybe),
                            toolz.map(onq.as_measurement(self.expect_project_units.LENGTH)),
                            list, )
        return WellHeadLocation(*result)

    def stages(self) -> oss.SearchableStages:
        """
        Return a `spo.SearchableProjectObjects` instance of all the stages for this project.

        Returns:
            An `spo.SearchableProjectObjects` for all the stages of this project.
        """
        return oss.SearchableStages(nsa.NativeStageAdapter, self.dom_object.Stages.Items)

    def locations_for_md_kb_values(self,
                                   md_kb_values: Iterable[om.Quantity],
                                   well_reference_frame_xy: origins.WellReferenceFrameXy,
                                   depth_origin: origins.DepthDatum) -> Iterable[nsp.SubsurfacePoint]:
        sample_at = Array[UnitsNet.Length](toolz.map(onq.as_net_quantity(self.expect_project_units.LENGTH),
                                                     md_kb_values))
        result = toolz.pipe(
            self.dom_object.GetLocationsForMdKbValues(sample_at, well_reference_frame_xy, depth_origin),
            toolz.map(nsp.make_subsurface_point(self.expect_project_units.LENGTH)),
            list,
        )
        return result

    def add_stages(self, create_stage_dtos: Iterable[nsa.CreateStageDto]):
        created_stages = [csd.create_stage(self) for csd in create_stage_dtos]

        with dnd.disposable(self.dom_object.ToMutable()) as mutable_well:
            native_created_stages = self._create_net_stages(created_stages)
            mutable_well.AddStages(native_created_stages)

    @staticmethod
    def _create_net_stages(created_stages):
        """
        Create a .NET `Array<IStage>`.

        This method primarily exists so that I can mock the call in unit tests.

        Args:
            created_stages: The `NativeStageAdapter` iterable over the created stages to add.

        Returns:
            The newly created .NET `Array<IStage>` instance.
        """
        return Array[IStage]([created_stage.dom_object for created_stage in created_stages])



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#


import pint


registry = pint.UnitRegistry()
"""
The single registry of all known `pint` units. See the Pint tutorial, 
https://pint.readthedocs.io/en/stable/tutorial.html, for general information on the registry. Specifically, 
see the section, https://pint.readthedocs.io/en/stable/tutorial.html#using-pint-in-your-projects, for the 
"perils" of using multiple registry instances.
"""

# Expose general types for use by type annotations
Quantity = registry.Quantity
"""The type of a Pint measurement exposed for convenience."""
Unit = registry.Unit
"""The type of Pint units of measure."""

# Register this instance of the registry as the application registry to support pickling and unpickling of Pint
# Quantity and Unit instances.
pint.set_application_registry(registry)



#! /usr/bin/env python
#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#


import argparse
import glob
import pathlib
import shutil
import site
import sys
from typing import Optional, Sequence


def site_packages_path() -> pathlib.Path:
    candidate_site_packages = [pn for pn in site.getsitepackages() if pn.find('site-packages') != -1]
    assert len(candidate_site_packages) == 1

    result = pathlib.Path(candidate_site_packages[0])
    return result


def copy_examples_to(target_dir: str, overwrite: bool) -> None:
    """
    Copy the Orchid Python API examples into the specified, `target_dir`.
    Args:
        target_dir: The target for the examples.
        overwrite: Flag: true if script will overwrite existing files with the same name in the target_dir.
    """

    def example_candidates(filename_glob):
        example_glob = str(site_packages_path().joinpath('orchid_python_api', 'examples', filename_glob))
        return glob.glob(example_glob), example_glob

    example_notebook_candidates, example_notebook_glob = example_candidates('*.ipynb')
    if not example_notebook_candidates:
        print(f'No example notebooks matching "{example_notebook_glob}"')
        return

    example_script_candidates, example_script_glob = example_candidates('*.py')
    if not example_script_candidates:
        print(f'No example scripts matching "{example_script_glob}"')
        return

    candidates = example_notebook_candidates + example_script_candidates
    for src in candidates:
        target_path = pathlib.Path(target_dir).joinpath(pathlib.Path(src).name)
        duplicate_in_target = target_path.exists()
        if duplicate_in_target and not overwrite:
            print(f'Skipping "{src}". Already exists.')
        else:
            shutil.copy2(src, target_dir)
            print(f'Copied "{src}" to "{target_dir}"')


def main(cli_args: Optional[Sequence[str]] = None):
    """
    Entry point for copy Orchid examples utility.
    Args:
        cli_args: The command line arguments.
    """
    cli_args = cli_args if cli_args else sys.argv[1:]

    parser = argparse.ArgumentParser(
        description='Copy Orchid Python API examples from installed package to specified directory')
    parser.add_argument('-t', '--target-dir', default='.',
                        help='Directory into which to copy the Orchid Python API examples '
                             '(default: current directory)')
    parser.add_argument('-o', '--overwrite', action='store_true', default=False,
                        help='Overwrite existing files in target directory',)

    args = parser.parse_args(cli_args)
    copy_examples_to(args.target_dir, args.overwrite)


if __name__ == '__main__':
    main(sys.argv[1:])



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#

from typing import Union

import toolz.curried as toolz

from orchid import (
    measurement as om,
    unit_system as units,
)


@toolz.curry
def to_unit(target_unit: Union[units.UsOilfield, units.Metric], source_measurement: om.Quantity):
    """
    Convert a `Measurement` instance to the same measurement in `target_unit`.

    The order of arguments allows easier conversion of a sequence of `Measurement` instances (with the same
    unit) to another unit. For example, if client code wished to convert a sequence of force measurements from
    US oilfield units to metric units (that is, pound-force to Newtons). Code to perform this conversion might
    be similar to the following:

    > make_metric_force = to_unit(units.Metric.FORCE)
    > metric_force_measurements = [make_metric_force(f) for f in us_oilfield_force_measurements]
    > # alternatively,
    > # metric_force_measurements = toolz.map(make_metric_force, us_oilfield_force_measurements)

    Args:
        source_measurement: The Measurement instance to convert.
        target_unit: The units to which I convert `source_measurement`.
    """
    return source_measurement.to(target_unit.value.unit)



#  Copyright (c) 2017-2022 Reveal Energy Services, Inc 
#
#  Licensed under the Apache License, Version 2.0 (the "License"); 
#  you may not use this file except in compliance with the License. 
#  You may obtain a copy of the License at 
#
#      http://www.apache.org/licenses/LICENSE-2.0 
#
#  Unless required by applicable law or agreed to in writing, software 
#  distributed under the License is distributed on an "AS IS" BASIS, 
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
#  See the License for the specific language governing permissions and 
#  limitations under the License. 
#
# This file is part of Orchid and related technologies.
#


import dataclasses as dc
from enum import IntEnum
import math
from typing import Optional, Tuple, Union

import deal
import option
import pendulum as pdt
import toolz.curried as toolz

import orchid.base
from orchid import (
    dot_net_disposable as dnd,
    dot_net_dom_access as dna,
    dom_project_object as dpo,
    measurement as om,
    native_stage_part_adapter as spa,
    native_subsurface_point as nsp,
    native_treatment_curve_adapter as ntc,
    net_date_time as ndt,
    net_fracture_diagnostics_factory as fdf,
    net_quantity as onq,
    reference_origins as origins,
    searchable_stage_parts as ssp,
    unit_system as units,
    validation
)

# noinspection PyUnresolvedReferences,PyPackageRequirements
from Orchid.FractureDiagnostics import FormationConnectionType, IStagePart
# noinspection PyUnresolvedReferences,PyPackageRequirements
from Orchid.FractureDiagnostics.Factories import Calculations
# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.SDKFacade import ScriptAdapter
# noinspection PyUnresolvedReferences
import UnitsNet
# noinspection PyUnresolvedReferences
from System.Collections.Generic import List
# noinspection PyUnresolvedReferences
import System


VALID_LENGTH_UNIT_MESSAGE = 'The parameter, `in_length_unit`, must be a unit system length.'


_object_factory = fdf.create()


class ConnectionType(IntEnum):
    PLUG_AND_PERF = FormationConnectionType.PlugAndPerf,
    SLIDING_SLEEVE = FormationConnectionType.SlidingSleeve,
    SINGLE_POINT_ENTRY = FormationConnectionType.SinglePointEntry,
    OPEN_HOLE = FormationConnectionType.OpenHole


def as_connection_type(type_value):
    @toolz.curry
    def has_value(value, enum_type):
        return value == enum_type.value

    return toolz.pipe(iter(ConnectionType),
                      toolz.filter(has_value(type_value)),
                      toolz.nth(0))


# This pre-condition applies to the public methods:
# - bottom_location
# - center_location
# - cluster_location
# - top_location
# but is implemented here to ensure common behavior among all these public methods.
@deal.pre(lambda _depth_datum, _reference_frame, in_length_unit, _net_subsurface_point_func:
          validation.is_unit_system_length(in_length_unit),
          message=VALID_LENGTH_UNIT_MESSAGE)
def subsurface_point_in_length_unit(depth_datum: origins.DepthDatum,
                                    xy_reference_frame: origins.WellReferenceFrameXy,
                                    in_length_unit: Union[units.UsOilfield, units.Metric],
                                    net_subsurface_point_func) -> nsp.SubsurfacePoint:
    """
    Calculate the subsurface point `in_length_unit` whose value is calculated by the
    callable, `net_subsurface_point_func`.

    Although this method is public, the author intends it to be "private." The author has made it public
    **only** to support unit testing. No other usage is supported.

    Args:
        depth_datum: The datum from which we measure depths.
        xy_reference_frame: The reference frame for easting-northing coordinates.
        in_length_unit: The unit of length available from the returned value.
        net_subsurface_point_func: The callable to calculate the subsurface point in .NET.

    Returns:
        The subsurface point in the requested unit of length.
    """
    net_subsurface_point = net_subsurface_point_func(xy_reference_frame.value, depth_datum.value)
    result = nsp.SubsurfacePoint(net_subsurface_point, in_length_unit)
    return result


class NativeStageAdapter(dpo.DomProjectObject):
    """Adapts a .NET IStage to be more Pythonic."""

    def __init__(self, adaptee, calculations_factory=None):
        super().__init__(adaptee, orchid.base.constantly(adaptee.Well.Project))
        self.calculations_factory = Calculations.FractureDiagnosticsCalculationsFactory() \
            if not calculations_factory else calculations_factory

    cluster_count = dna.dom_property('number_of_clusters', 'The number of clusters for this stage')
    display_name_with_well = dna.dom_property('display_name_with_well',
                                              'The display stage number including the well name')
    display_name_without_well = dna.dom_property('display_name_without_well',
                                                 'The display stage number excluding the well name')
    display_stage_number = dna.dom_property('display_stage_number', 'The display stage number for the stage')
    global_stage_sequence_number = dna.dom_property('global_stage_sequence_number',
                                                    'The global sequence number of this stage')
    order_of_completion_on_well = dna.dom_property('order_of_completion_on_well',
                                                   'The order in which this stage was completed on its well')
    stage_type = dna.transformed_dom_property('stage_type', 'The formation connection type of this stage',
                                              as_connection_type)
    start_time = dna.transformed_dom_property('start_time', 'The start time of the stage treatment',
                                              ndt.as_date_time)
    stop_time = dna.transformed_dom_property('stop_time', 'The stop time of the stage treatment',
                                             ndt.as_date_time)

    def _get_time_range(self) -> pdt.Period:
        return pdt.period(self.start_time, self.stop_time)

    def _set_time_range(self, to_time_range: pdt.Period):
        to_start_net_time = ndt.as_net_date_time(to_time_range.start)
        to_stop_net_time = ndt.as_net_date_time(to_time_range.end)
        if len(self.stage_parts()) == 1:
            single_stage_part = toolz.first(self.stage_parts())
            with dnd.disposable(single_stage_part.dom_object.ToMutable()) as mutable_first_stage_part:
                mutable_first_stage_part.SetStartStopTimes(to_start_net_time,
                                                           to_stop_net_time)
        elif len(self.stage_parts()) > 1:
            single_stage_part = toolz.first(self.stage_parts())
            with dnd.disposable(single_stage_part.dom_object.ToMutable()) as mutable_first_stage_part:
                mutable_first_stage_part.SetStartStopTimes(to_start_net_time,
                                                           single_stage_part.dom_object.StopTime)
            last_stage_part = toolz.last(self.stage_parts())
            with dnd.disposable(last_stage_part.dom_object.ToMutable()) as mutable_last_stage_part:
                mutable_last_stage_part.SetStartStopTimes(last_stage_part.dom_object.StartTime,
                                                          to_stop_net_time)
        else:  # No stage parts
            factory = fdf.create()
            stage_part_to_add = factory.CreateStagePart(self.dom_object, to_start_net_time,
                                                        to_stop_net_time, None)
            with dnd.disposable(self.dom_object.ToMutable()) as mutable_stage:
                mutable_stage.Parts.Add(stage_part_to_add)

    time_range = property(fget=_get_time_range, fset=_set_time_range,
                          doc='The time range (start and end) of this stage')

    @property
    def isip(self) -> om.Quantity:
        """
        Return the instantaneous shut in pressure of this stage in project units.
        """
        return onq.as_measurement_from_option(self.expect_project_units.PRESSURE, self.dom_object.Isip)

    @property
    def pnet(self) -> om.Quantity:
        """
        Return the net pressure of this stage in project units.

        The net pressure of a stage is calculated by the formula:
            pnet = isip + fluid-density * tvd - shmin (where tvd is the true vertical depth)
        """
        return onq.as_measurement_from_option(self.expect_project_units.PRESSURE, self.dom_object.Pnet)

    @property
    def shmin(self) -> om.Quantity:
        """
        Return the minimum horizontal stress of this stage in project units.
        """

        return onq.as_measurement_from_option(self.expect_project_units.PRESSURE, self.dom_object.Shmin)

    @staticmethod
    def _sampled_quantity_name_curve_map(sampled_quantity_name):
        candidates = toolz.pipe(ntc.TreatmentCurveTypes,
                                toolz.filter(lambda e: e.value == sampled_quantity_name),
                                list)
        if len(candidates) == 0:
            raise KeyError(f'Unknown sampled quantity name: "{sampled_quantity_name}"')

        assert len(candidates) == 1, f'Sampled quantity name "{sampled_quantity_name}"' \
                                     f' selects many curve types: {candidates}'

        return candidates[0]

    def _center_location_depth(self, in_length_unit: Union[units.UsOilfield, units.Metric],
                               depth_datum: origins.DepthDatum) -> om.Quantity:
        """
        Return the depth of the stage center relative to the specified `depth_datum.`

        Args:
            in_length_unit: The unit of length for the returned Measurement.
            depth_datum: The reference datum for the depth.
        """
        subsurface_point = self.center_location(in_length_unit, origins.WellReferenceFrameXy.ABSOLUTE_STATE_PLANE,
                                                depth_datum)
        return subsurface_point.depth

    def bottom_location(self, in_length_unit: Union[units.UsOilfield, units.Metric],
                        xy_reference_frame: origins.WellReferenceFrameXy,
                        depth_datum: origins.DepthDatum) -> nsp.SubsurfacePoint:
        """
        Return the location of the bottom of this stage in the `xy_well_reference_frame` using the
        `depth_datum` in the specified unit.

        Args:
            in_length_unit: The unit of length available from the returned value.
            xy_reference_frame: The reference frame for easting-northing coordinates.
            depth_datum: The datum from which we measure depths.

        Returns:
            The `SubsurfacePoint` of the stage bottom.
        """

        return subsurface_point_in_length_unit(depth_datum, xy_reference_frame, in_length_unit,
                                               self.dom_object.GetStageLocationBottom)

    def center_location(self, in_length_unit: Union[units.UsOilfield, units.Metric],
                        xy_reference_frame: origins.WellReferenceFrameXy,
                        depth_datum: origins.DepthDatum) -> nsp.SubsurfacePoint:
        """
        Return the location of the center of this stage in the `xy_well_reference_frame` using the `depth_datum`
        in the specified unit.

        Args:
            in_length_unit: The unit of length available from the returned value.
            xy_reference_frame: The reference frame for easting-northing coordinates.
            depth_datum: The datum from which we measure depths.

        Returns:
            The `SubsurfacePoint` of the stage center.
        """
        return subsurface_point_in_length_unit(depth_datum, xy_reference_frame, in_length_unit,
                                               self.dom_object.GetStageLocationCenter)

    def center_location_easting(self, in_length_unit: Union[units.UsOilfield, units.Metric],
                                xy_well_reference_frame: origins.WellReferenceFrameXy) -> om.Quantity:
        """
        Return the easting location of the stage center relative to the specified reference frame in the
        specified unit.

        Args:
            in_length_unit: An unit of the unit of length for the returned Measurement.
            xy_well_reference_frame: The reference frame defining the origin.

        Returns:
            A measurement.
        """
        result = self.center_location(in_length_unit, xy_well_reference_frame, origins.DepthDatum.KELLY_BUSHING).x
        return result

    def center_location_northing(self, in_length_unit: Union[units.UsOilfield, units.Metric],
                                 xy_well_reference_frame: origins.WellReferenceFrameXy) -> om.Quantity:
        """
        Return the northing location of the stage center in the `xy_well_reference_frame` in the specified unit.

        Args:
            in_length_unit: The requested resultant length unit.
            xy_well_reference_frame: The reference frame defining the origin.

        Returns:
            A measurement.
        """
        subsurface_point = self.center_location(in_length_unit, xy_well_reference_frame,
                                                origins.DepthDatum.KELLY_BUSHING)
        return subsurface_point.y

    def center_location_mdkb(self, in_length_unit: Union[units.UsOilfield, units.Metric]) -> om.Quantity:
        """
        Return the measured depth of the stage center in project units.

        Args:
            in_length_unit: The unit of length for the returned Measurement.
        """
        return (self.md_top(in_length_unit) + self.md_bottom(in_length_unit)) / 2

    def center_location_tvdgl(self, in_length_unit: Union[units.UsOilfield, units.Metric]) -> om.Quantity:
        """
        Returns the total vertical depth from ground level of the stage center in project units.

        Args:
            in_length_unit: The unit of length for the returned Measurement.
        """
        return self._center_location_depth(in_length_unit, origins.DepthDatum.GROUND_LEVEL)

    def center_location_tvdss(self, in_length_unit: Union[units.UsOilfield, units.Metric]) -> om.Quantity:
        """
        Returns the total vertical depth from sea level of the stage center in project units.

        Args:
            in_length_unit: The unit of length for the returned Measurement.
        """
        return self._center_location_depth(in_length_unit, origins.DepthDatum.SEA_LEVEL)

    def center_location_xy(self, in_length_unit: Union[units.UsOilfield, units.Metric],
                           xy_well_reference_frame: origins.WellReferenceFrameXy) -> Tuple[om.Quantity,
                                                                                           om.Quantity]:
        """
        Return the easting-northing location of the stage center in the `xy_well_reference_frame` in project units.

        Args:
            in_length_unit: The unit of length for the returned Measurement.
            xy_well_reference_frame: The reference frame defining the origin.

        Returns:
            A tuple
        """
        subsurface_point = self.center_location(in_length_unit, xy_well_reference_frame,
                                                origins.DepthDatum.KELLY_BUSHING)
        return subsurface_point.x, subsurface_point.y

    @deal.pre(lambda _self, _in_length_unit, cluster_no, _xy_reference_frame, _depth_datum: cluster_no >= 0)
    def cluster_location(self, in_length_unit: Union[units.UsOilfield, units.Metric], cluster_no: int,
                         xy_reference_frame: origins.WellReferenceFrameXy,
                         depth_datum: origins.DepthDatum) -> nsp.SubsurfacePoint:
        """
        Return the location of the bottom of this stage in the `xy_well_reference_frame` using the
        `depth_datum` in the specified unit.

        Args:
            in_length_unit: The unit of length available from the returned value.
            cluster_no: The number identifying the cluster whose location is sought.
            xy_reference_frame: The reference frame for easting-northing coordinates.
            depth_datum: The datum from which we measure depths.

        Returns:
            The `SubsurfacePoint` of the stage cluster identified by `cluster_no`.
        """
        stage_location_cluster_func = toolz.curry(self.dom_object.GetStageLocationCluster, cluster_no)
        return subsurface_point_in_length_unit(depth_datum, xy_reference_frame, in_length_unit,
                                               stage_location_cluster_func)

    @deal.pre(validation.arg_is_acceptable_pressure_unit)
    def isip_in_pressure_unit(self, target_unit: Union[units.UsOilfield, units.Metric]) -> om.Quantity:
        return onq.as_measurement_from_option(target_unit, self.dom_object.Isip)

    def md_bottom(self, in_length_unit: Union[units.UsOilfield, units.Metric]):
        """
        Return the measured depth of the bottom of this stage (farthest from the wellhead / closest to the toe)
        in the specified unit.

        Args:
            in_length_unit: An unit of the unit of length for the returned Measurement.

        Returns:
             The measured depth of the stage bottom in the specified unit.
        """
        return onq.as_measurement(in_length_unit, option.maybe(self.dom_object.MdBottom))

    def md_top(self, in_length_unit: Union[units.UsOilfield, units.Metric]) -> om.Quantity:
        """
        Return the measured depth of the top of this stage (closest to the wellhead / farthest from the toe)
        in the specified unit.

        Args:
            in_length_unit: An unit of the requested resultant length unit.

        Returns;
         The measured depth of the stage top in the specified unit.
        """
        return onq.as_measurement(in_length_unit, option.maybe(self.dom_object.MdTop))

    @deal.pre(validation.arg_is_acceptable_pressure_unit)
    def pnet_in_pressure_unit(self, target_unit: Union[units.UsOilfield, units.Metric]) -> om.Quantity:
        return onq.as_measurement_from_option(target_unit, self.dom_object.Pnet)

    @deal.pre(validation.arg_is_acceptable_pressure_unit)
    def shmin_in_pressure_unit(self, target_unit: Union[units.UsOilfield, units.Metric]) -> om.Quantity:
        return onq.as_measurement_from_option(target_unit, self.dom_object.Shmin)

    def stage_length(self, in_length_unit: Union[units.UsOilfield, units.Metric]) -> om.Quantity:
        """
        Return the stage length in the specified unit.

        Args:
            in_length_unit: An unit of the unit of length for the returned Measurement.

        Returns:
            The Measurement of the length of this stage.
        """
        return self.md_bottom(in_length_unit) - self.md_top(in_length_unit)

    def stage_parts(self) -> ssp.SearchableStageParts:
        """
        Return a `ssp.SearchableStageParts` for all the stage parts for this stage.

        Returns:
            An `ssp.SearchableStageParts` for all the stage parts for this stage.
        """
        return ssp.SearchableStageParts(spa.NativeStagePartAdapter, self.dom_object.Parts)

    def top_location(self, in_length_unit: Union[units.UsOilfield, units.Metric],
                     xy_reference_frame: origins.WellReferenceFrameXy,
                     depth_datum: origins.DepthDatum) -> nsp.SubsurfacePoint:
        """
        Return the location of the top of this stage in the `xy_well_reference_frame` using the `depth_datum`
        in the specified unit.

        Args:
            in_length_unit: The unit of length available from the returned value.
            xy_reference_frame: The reference frame for easting-northing coordinates.
            depth_datum: The datum from which we measure depths.

        Returns:
            The `SubsurfacePoint` of the stage top.
        """
        return subsurface_point_in_length_unit(depth_datum, xy_reference_frame, in_length_unit,
                                               self.dom_object.GetStageLocationTop)

    def treatment_curves(self):
        """
        Returns the dictionary of treatment curves for this treatment_stage.

        Request a specific curve from the dictionary using the constants defined in `orchid`:

        - `PROPPANT_CONCENTRATION`
        - `SLURRY_RATE`
        - `TREATING_PRESSURE`

        Returns:
            The dictionary containing the available treatment curves.
        """
        if not self.dom_object.TreatmentCurves.Items:
            return {}

        def add_curve(so_far, treatment_curve):
            curve_name = self._sampled_quantity_name_curve_map(treatment_curve.sampled_quantity_name)
            treatment_curve_map = {curve_name: treatment_curve}
            return toolz.merge(treatment_curve_map, so_far)

        result = toolz.pipe(self.dom_object.TreatmentCurves.Items,  # start with .NET treatment curves
                            toolz.map(ntc.NativeTreatmentCurveAdapter),  # wrap them in a facade
                            # Transform the map to a dictionary keyed by the sampled quantity name
                            lambda cs: toolz.reduce(add_curve, cs, {}))
        return result


@dc.dataclass
class CreateStageDto:
    """
    A data-transfer object (DTO) containing the required and optional data used to create new stages.

    This class enforces a number of constraints on the data needed to create a stage. Consider viewing the source
    code of the class to understand those constraints. (The class will throw exceptions if the constraints are not met
    at run-time.)

    Additionally, the comments of this class contain a number of "warnings" for situations where data is not required,
    but the consequences for **not** supplying the data may not be desirable or may not be expected.
    """
    stage_no: int  # Must be positive
    connection_type: ConnectionType
    md_top: om.Quantity  # Must be length
    md_bottom: om.Quantity  # Must be length

    cluster_count: int = 0  # Must be non-negative
    maybe_shmin: Optional[om.Quantity] = None  # If not `None`, must be pressure
    # WARNING: one need supply neither a start time nor a stop time; however, not supplying this data can
    # produce unexpected behavior for the `global_stage_sequence_number` property. For example, one can
    # generate duplicate values for the `global_stage_sequence_number`. This unexpected behavior is a known
    # issue with Orchid.
    #
    # Note supplying no value (an implicit `None`) results in the largest possible .NET time range.
    maybe_time_range: Optional[pdt.Period] = None
    # WARNING: one must currently supply an ISIP for each stage; otherwise, Orchid fails to correctly load
    # the project saved with the added stages.
    maybe_isip: Optional[om.Quantity] = None  # If not `None`, must be a pressure

    order_of_completion_on_well = property(fget=lambda self: self.stage_no - 1)

    def __post_init__(self):
        # See the
        # [StackOverflow post](https://stackoverflow.com/questions/54488765/validating-input-when-mutating-a-dataclass)
        if self.stage_no <= 0:
            raise ValueError(f'Expected stage_no to be positive. Found {self.stage_no}')
        if not self.md_top.check('[length]'):
            raise ValueError(f'Expected md_top to be a length. Found {self.md_top:~P}')
        if not self.md_bottom.check('[length]'):
            raise ValueError(f'Expected md_bottom to be a length. Found {self.md_bottom:~P}')
        if self.cluster_count < 0:
            raise ValueError(f'Expected cluster_count to be non-negative. Found {self.cluster_count}')
        if self.maybe_isip is not None:
            if not self.maybe_isip.check('[pressure]'):
                raise ValueError(f'Expected maybe_isip to be a pressure if not None. Found {self.maybe_isip:~P}')
        if self.maybe_shmin is not None:
            if not self.maybe_shmin.check('[pressure]'):
                raise ValueError(f'Expected maybe_shmin to be a pressure if not None. Found {self.maybe_shmin:~P}')

    def create_stage(self, well) -> NativeStageAdapter:
        """
        Creates a stage from this DTO.

        Args:
            well (NativeWellAdapter): The well of the created `NativeStageAdapter`

        Returns:
            The `NativeStageAdapter` wrapping the created .NET `IStage` instance.

        """
        project_unit_system = units.as_unit_system(well.dom_object.Project.ProjectUnits)
        native_md_top = onq.as_net_quantity(project_unit_system.LENGTH, self.md_top)
        native_md_bottom = onq.as_net_quantity(project_unit_system.LENGTH, self.md_bottom)
        if self.maybe_shmin is None:
            native_shmin = ScriptAdapter.MakeOptionNone[UnitsNet.Pressure]()
        elif math.isnan(self.maybe_shmin.magnitude):
            native_shmin = ScriptAdapter.MakeOptionNone[UnitsNet.Pressure]()
        else:
            native_shmin = ScriptAdapter.MakeOptionSome(
                onq.as_net_quantity(project_unit_system.PRESSURE, self.maybe_shmin))
        completion_order_on_well = System.UInt32(self.order_of_completion_on_well)
        connection_type = self.connection_type.value
        cluster_count = System.UInt32(self.cluster_count)
        no_time_range_native_stage = self.create_net_stage(well.dom_object, completion_order_on_well,
                                                           connection_type, native_md_top,
                                                           native_md_bottom, native_shmin,
                                                           cluster_count)

        with dnd.disposable(no_time_range_native_stage.ToMutable()) as mutable_stage:
            native_start_time = (ndt.as_net_date_time(self.maybe_time_range.start)
                                 if self.maybe_time_range is not None
                                 else pdt.DateTime.max)
            native_stop_time = (ndt.as_net_date_time(self.maybe_time_range.end)
                                if self.maybe_time_range is not None
                                else pdt.DateTime.max)
            if self.maybe_isip is None:
                native_isip = None
            elif math.isnan(self.maybe_isip.magnitude):
                native_isip = None
            else:
                native_isip = onq.as_net_quantity(project_unit_system.PRESSURE, self.maybe_isip)

            stage_part = self.create_net_stage_part(no_time_range_native_stage, native_start_time, native_stop_time,
                                                    native_isip)
            self.add_stage_part_to_stage(mutable_stage, stage_part)

        # Alias to better communicate intent
        created_native_stage = no_time_range_native_stage
        return NativeStageAdapter(created_native_stage)

    @staticmethod
    def create_net_stage(native_well, completion_order_on_well, connection_type, native_md_top, native_md_bottom,
                         native_shmin, cluster_count):
        """
        Create a .NET `IStage`.

        This method primarily exists so that I can mock the call in unit tests.

        Args:
            native_well: The .NET `IWell` instance to which the created `IStage` refers.
            completion_order_on_well: The order of completion of this stage of the referenced `IWell`.
            connection_type: The .NET `FormationConnectionType` of the created .NET stage.
            native_md_top: The measured depth of the top (toward the heel) of the created .NET stage.
            native_md_bottom: The measured depth of the bottom (toward the toe) of the created .NET stage.
            native_shmin: The minimum horizontal stress of the created .NET stage.
            cluster_count: The cluster count of the created .NET stage.

        Returns:
            The newly created .NET `IStage` instance.
        """
        no_time_range_native_stage = _object_factory.CreateStage(
            System.UInt32(completion_order_on_well),
            native_well,
            connection_type,
            native_md_top,
            native_md_bottom,
            native_shmin,
            System.UInt32(cluster_count)
        )
        return no_time_range_native_stage

    @staticmethod
    def create_net_stage_part(native_stage, native_start_time, native_stop_time, native_isip):
        """
        Create . .NET `IStagePart` instance.

        This method primarily exists so that I can mock the call in unit tests.

        Args:
            native_stage: The .NET `IStage` to which the created `IStagePart` refers.
            native_start_time: The start time of the created `IStagePart`.
            native_stop_time: The stop time of the created `IStagePart`.
            native_isip: The ISIP of the created `IStagePart`.

        Returns:
            The newly created `IStagePart` with the specified details.
        """
        stage_part = _object_factory.CreateStagePart(native_stage,
                                                     native_start_time,
                                                     native_stop_time,
                                                     native_isip)
        return stage_part

    @staticmethod
    def add_stage_part_to_stage(mutable_stage, stage_part):
        """
        Add a newly created `stage_part` to a newly created `mutable_stage`.

        This method exists primarily so that I can mock it for unit tests.

        Args:
            mutable_stage: The newly created stage supporting mutability.
            stage_part: The newly created stage part.
        """
        stage_parts = List[IStagePart]()
        stage_parts.Add(stage_part)
        mutable_stage.Parts = stage_parts



#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#

import pathlib
import pprint
from uuid import UUID

import pandas as pd

import orchid

# Load the Orchid project
default_file_name_to_read = pathlib.Path('frankNstein_Bakken_UTM13_FEET.ifrac')
default_project_path_name_to_read = str(orchid.training_data_path().joinpath(default_file_name_to_read))
project = orchid.load_project(default_project_path_name_to_read)

# Get a dictionary mapping object IDs to project time series
time_series = {ts.object_id: ts for ts in project.time_series().all_objects()}
print('\nAll time series in project')
pprint.pprint(time_series)

# Get a .NET (native) monitor whose time series you want to use
monitors = {m.object_id: m for m in project.monitors().all_objects()}
print('\nAll monitors in project')
pprint.pprint(monitors)

# In this particular scenario, I'm simply going to pick the first monitor from a dictionary mapping object IDs
# to native_monitor_adapters.
monitor_oid = UUID('5b68d8c4-a578-44e7-bc08-b1d83483c4ec')
monitor = monitors[monitor_oid]
print('\nMonitor of interest:')
print(f'  - Object ID: {monitor.object_id}')
print(f'  - Display Name: {monitor.display_name}')

# Extract the object ID of the time series associated with this monitor
monitor_time_series_native_oid = monitors[monitor_oid].dom_object.TimeSeries.ObjectId
monitor_time_series_oid = UUID(str(monitor_time_series_native_oid))
print(f'\nObject ID of monitor time series of interest: {monitor_time_series_oid}')

# And get the time series identified by this object ID
time_series_for_monitor = project.time_series().find_by_object_id(monitor_time_series_oid)

# Finally, get the time series as a `pandas` (time) `Series`
time_series_of_interest = time_series_for_monitor.data_points()
print('\nHead of time series')
print(time_series_of_interest.head())




#
# This file is part of Orchid and related technologies.
#
# Copyright (c) 2017-2022 Reveal Energy Services.  All Rights Reserved.
#
# LEGAL NOTICE:
# Orchid contains trade secrets and otherwise confidential information
# owned by Reveal Energy Services. Access to and use of this information is 
# strictly limited and controlled by the Company. This file may not be copied,
# distributed, or otherwise disclosed outside of the Company's facilities 
# except under appropriate precautions to maintain the confidentiality hereof, 
# and may not be used in any way not expressly authorized by the Company.
#


from abc import ABCMeta, abstractmethod
from typing import Callable, Union

import numpy as np
import pandas as pd

from orchid import (
    dom_project_object as dpo,
    dot_net_dom_access as dna,
    project_store as loader,
    unit_system as units,
)

# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.SDKFacade import ScriptAdapter
# noinspection PyUnresolvedReferences
from Orchid.FractureDiagnostics.TimeSeries import IQuantityTimeSeries


class BaseTimeSeriesAdapter(dpo.DomProjectObject, metaclass=ABCMeta):
    def __init__(self, adaptee: IQuantityTimeSeries, net_project_callable: Callable):
        """
        Construct an instance that adapts a .NET `IStageSampledQuantityTimeSeries` instance.

        Args:
            adaptee: The .NET stage time series to be adapted.
        """
        super().__init__(adaptee, net_project_callable)

    sampled_quantity_name = dna.dom_property('sampled_quantity_name',
                                             'Return the sampled quantity name for this curve.')

    @abstractmethod
    def quantity_name_unit_map(self, project_units):
        """
        Return a map (dictionary) between quantity names and units (from `unit_system`) of the data_points.

        This method plays the role of "Primitive Operation" in the *Template Method* design pattern. In this
        role, the "Template Method" defines an algorithm and delegates some steps of the algorithm to derived
        classes through invocation of "Primitive Operations".

        Args:
            project_units: The unit system of the project.
        """
        pass

    def sampled_quantity_unit(self) -> Union[units.UsOilfield, units.Metric]:
        """
        Return the measurement unit of the data_points in this curve.

        This method plays the role of "Template Method" in the *Template Method* design pattern. In this role
        it specifies an algorithm to calculate the units of the sampled quantity of the curve delegating some
        algorithm steps to derived classes by invoking the "Primitive Operation-", `quantity_name_unit_map()`
        and `get_net_project_units()`.

        Returns:
            A `UnitSystem` member containing the unit for the sample in this curve.
        """
        quantity_name_unit_map = self.quantity_name_unit_map(self.expect_project_units)
        return quantity_name_unit_map[self.sampled_quantity_name]

    def data_points(self) -> pd.Series:
        """
        Return the time series for this curve.

        Returns
            The `pandas` time `Series` for this curve.
        """
        python_time_series_arrays = loader.as_python_time_series_arrays(self.dom_object)

        result = pd.Series(data=np.fromiter(python_time_series_arrays.SampleMagnitudes, dtype='float'),
                           index=pd.DatetimeIndex(np.fromiter(python_time_series_arrays.UnixTimeStampsInSeconds,
                                                              dtype='datetime64[s]'), tz='UTC'),
                           name=self.name)

        return result
